# Copyright French Alternative Energies and Atomic Energy Commission
# Contributors: valjean developers
# valjean-support@cea.fr
#
# This software is a computer program whose purpose is to analyze and
# post-process numerical simulation results.
#
# This software is governed by the CeCILL license under French law and abiding
# by the rules of distribution of free software. You can use, modify and/ or
# redistribute the software under the terms of the CeCILL license as circulated
# by CEA, CNRS and INRIA at the following URL: http://www.cecill.info.
#
# As a counterpart to the access to the source code and rights to copy, modify
# and redistribute granted by the license, users are provided only with a
# limited warranty and the software's author, the holder of the economic
# rights, and the successive licensors have only limited liability.
#
# In this respect, the user's attention is drawn to the risks associated with
# loading, using, modifying and/or developing or reproducing the software by
# the user in light of its specific status of free software, that may mean that
# it is complicated to manipulate, and that also therefore means that it is
# reserved for developers and experienced professionals having in-depth
# computer knowledge. Users are therefore encouraged to load and test the
# software's suitability as regards their requirements in conditions enabling
# the security of their systems and/or data to be ensured and, more generally,
# to use and operate it in the same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.
#
# -*- coding: utf-8 -*-

'''This module contains an implementation of a scheduling backend that
leverages Python "threads" (:mod:`~threading` module) and producer-consumer
queues (:mod:`~queue` module).
'''

import threading
import logging
import time
from queue import Queue
from functools import partial

from ..task import TaskStatus
from ...chrono import Chrono


LOGGER = logging.getLogger(__name__)


class QueueScheduling:
    '''The default scheduling backend.

    Uses :mod:`threading` and :mod:`queue` to handle concurrent execution of
    tasks.
    '''

    def __init__(self, n_workers=10):

        '''Initialize the queue backend.

        :param int n_workers: The number of worker threads to use.
        '''
        self.n_workers = n_workers
        self.queue = Queue(0)

    def _enqueue(self, tasks, full_graph, hard_graph, env):
        '''Enqueue as many tasks as the dependencies permit.'''

        n_tasks = len(tasks)
        tasks_left = []
        LOGGER.debug('master: %d tasks left to schedule', n_tasks)
        for task in tasks:
            if LOGGER.isEnabledFor(logging.DEBUG):
                LOGGER.debug('master: considering task %s for scheduling: %s',
                             task, env.get_status(task))

            deps = full_graph.dependencies(task)
            hard_deps = hard_graph.dependencies(task)

            new_state = env.atomically(partial(self.decide_new_state,
                                               task, deps, hard_deps))

            if new_state == TaskStatus.SKIPPED:
                LOGGER.debug('master: task %s has failed deps', task)
                n_tasks -= 1
            elif new_state == TaskStatus.PENDING:
                # we start to work on the task
                self.queue.put(task)
                n_tasks -= 1
            elif new_state == TaskStatus.WAITING:
                LOGGER.debug('master: task %s blocked by dependencies', task)
                tasks_left.append(task)
            elif new_state is None:
                # intentionally leave the task out of the queue
                pass

        return tasks_left

    @classmethod
    def decide_new_state(cls, task, deps, hard_deps, env):
        '''Look at the dependencies of the current task and make a decision
        about what we should do with this task.

        This function returns a :class:`~.TaskStatus` that represents the
        suggested new state for this task, or `None` in case the task should be
        ignored and removed from the queue.

        :param Task task: the task
        :param list(Task) deps: the dependencies of `task`
        :param list(Task) hard_deps: the hard dependencies of `task`
        :param Env env: the environment.
        :rtype: TaskStatus or None
        '''
        LOGGER.debug('env in decide: %s', env)
        if any(t.name not in env or env.is_pending(t) for t in deps):
            LOGGER.debug('some of the dependencies of task %s are missing or '
                         'pending; waiting', task)
            env.set_waiting(task)
            return TaskStatus.WAITING

        if env.is_done(task):
            last_dep_end = cls.last_end_time(deps, env)
            if last_dep_end is None:
                LOGGER.info('task %s has no dependencies and is already done; '
                            'removing from queue', task)
                return None
            task_start = env.get_start_clock(task)
            if task_start is not None and last_dep_end <= task_start:
                LOGGER.info('task %s is newer than its dependencies and is '
                            'already done; removing from queue', task)
                return None
            LOGGER.info('task %s is already done but some if its dependencies '
                        'are newer; scheduling', task)
            env.set_pending(task)
            return TaskStatus.PENDING

        assert env.is_waiting(task)
        return cls.decide_new_state_waiting(task, deps, hard_deps, env)

    @classmethod
    def last_end_time(cls, tasks, env):
        '''Return the latest ``end_clock`` time of the given tasks, or `None`
        if some of the tasks do not have an ``end_clock``.'''
        ends = []
        for task in tasks:
            task_end = env.get_end_clock(task)
            if task_end is None:
                return None
            ends.append(task_end)
        return max(ends, default=None)

    @classmethod
    def decide_new_state_waiting(cls, task, deps, hard_deps, env):
        '''Decide what to do with a task that is in `WAITING` state.'''
        if any(env.is_failed(t) or env.is_skipped(t) for t in hard_deps):
            LOGGER.info('some of the hard dependencies of task %s failed; '
                        'skipping', task)
            env.set_skipped(task)
            return TaskStatus.SKIPPED

        if all(env.is_done(t) or env.is_failed(t) or env.is_skipped(t)
               for t in deps):
            LOGGER.debug('all the dependencies of task %s are done; '
                         'scheduling', task)
            env.set_pending(task)
            return TaskStatus.PENDING

        env.set_waiting(task)
        return TaskStatus.WAITING

    def execute_tasks(self, *, full_graph, hard_graph, env, config):
        '''Execute the tasks.

        :param DepGraph full_graph: Full dependency graph for the tasks, i.e.
                                    including both hard and soft dependencies.
        :param DepGraph hard_graph: Hard-dependency graph for the tasks.
        :param env: An initial environment for the scheduled tasks.
        :type env: Env
        :param env: The configuration object (for things like paths, etc.).
        :type env: Config
        '''

        threads = []
        # Use a condition variable to block when no more tasks can be submitted
        # because of dependencies. The master calls cond_var.wait() and the
        # worker threads call cond_var.notify() when they are finished with a
        # task.
        cond_var = threading.Condition()

        # spawn workers
        LOGGER.debug('master: spawning %s workers...', self.n_workers)
        for _ in range(self.n_workers):
            thread = QueueScheduling.WorkerThread(self.queue, env, config,
                                                  cond_var)
            thread.start()
            threads.append(thread)

        # process tasks; sort them in topological order
        with Chrono() as chrono:
            tasks_left = full_graph.topological_sort()
        LOGGER.info('full graph sorted in %s seconds', chrono)
        while tasks_left:
            n_tasks_left = len(tasks_left)
            LOGGER.info('master: %d tasks left', n_tasks_left)
            with cond_var:
                tasks_left = self._enqueue(tasks_left, full_graph, hard_graph,
                                           env)

                if not tasks_left:
                    # we're done!
                    break

                if n_tasks_left == len(tasks_left):
                    LOGGER.debug('master: worker queue is starving because of '
                                 'dependencies, going to sleep...')
                    # release the cond_var lock and wait for workers to call
                    # cond_var.notify()
                    cond_var.wait()
                    LOGGER.debug('master: ...woke up!')

        # block until all tasks are done
        self.queue.join()

        # stop workers and go home
        for _ in range(self.n_workers):
            self.queue.put(None)
        for thread in threads:
            thread.join()

    class WorkerThread(threading.Thread):
        '''Workhorse class for :class:`QueueScheduling`. This class consumes
        (i.e. executes) tasks passed to it through the queue.
        '''

        def __init__(self, queue, env, config, cond_var):
            '''Initialize the thread.

            :param queue: The producer-consumer task queue.
            :param env: The execution environment for the tasks.
            :param config: The configuration for the tasks.
            :param cond_var: A condition variable to notify when we are
                             finished running a task
            '''
            super().__init__()
            self.queue = queue
            self.config = config
            self.env = env
            self.cond_var = cond_var

        def run(self):
            '''Main method: run this thread.'''

            LOGGER.debug('worker %s: starting', self.name)
            while True:
                task = self.queue.get()
                if task is None:
                    LOGGER.debug('worker %s: exiting', self.name)
                    break
                LOGGER.debug('worker %s: got task %s', self.name, task)
                LOGGER.note('task %s starts', task)

                start = time.time()
                try:
                    task_result = task.do(self.env, self.config)
                except Exception as ex:  # pylint: disable=broad-except
                    LOGGER.exception('task %s on worker %s failed with the '
                                     'following exception:\n%s',
                                     task, self.name, ex)
                    LOGGER.debug('worker %s: setting FAILED status in the '
                                 'environment...', self.name)
                    self.env.set_failed(task)
                else:
                    if task_result is None:
                        LOGGER.error('expected an (env_update, status) pair '
                                     'from task %s, got None', task.name)
                        self.env.set_failed(task)
                    else:
                        env_update, status = task_result
                        LOGGER.debug('worker %s: task %s completed '
                                     'with status %s', self.name, task, status)
                        LOGGER.note('task %s completed with status %s',
                                    task, status)
                        LOGGER.debug('worker %s: proposed environment update: '
                                     '%s', self.name, env_update)
                        self.env.set_status(task, status)
                        self.env.apply(env_update)
                end = time.time()
                self.env.set_start_end_clock(task, start=start, end=end)

                LOGGER.debug('worker %s: wants more!', self.name)
                self.queue.task_done()

                # notify the master that a new task has completed
                with self.cond_var:
                    LOGGER.debug('worker %s: notifies master', self.name)
                    self.cond_var.notify_all()



# Copyright French Alternative Energies and Atomic Energy Commission
# Contributors: valjean developers
# valjean-support@cea.fr
#
# This software is a computer program whose purpose is to analyze and
# post-process numerical simulation results.
#
# This software is governed by the CeCILL license under French law and abiding
# by the rules of distribution of free software. You can use, modify and/ or
# redistribute the software under the terms of the CeCILL license as circulated
# by CEA, CNRS and INRIA at the following URL: http://www.cecill.info.
#
# As a counterpart to the access to the source code and rights to copy, modify
# and redistribute granted by the license, users are provided only with a
# limited warranty and the software's author, the holder of the economic
# rights, and the successive licensors have only limited liability.
#
# In this respect, the user's attention is drawn to the risks associated with
# loading, using, modifying and/or developing or reproducing the software by
# the user in light of its specific status of free software, that may mean that
# it is complicated to manipulate, and that also therefore means that it is
# reserved for developers and experienced professionals having in-depth
# computer knowledge. Users are therefore encouraged to load and test the
# software's suitability as regards their requirements in conditions enabling
# the security of their systems and/or data to be ensured and, more generally,
# to use and operate it in the same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

'''This module provides the classes to convert test results to reStructuredText
format. reStructuredText_ is a text markup language meant for easy human
consumption. The `reStructuredText primer`_ provides a good introduction to its
syntax.

.. _reStructuredText: https://docutils.sourceforge.io/rst.html
.. _reStructuredText primer: https://
    docutils.sourceforge.io/docs/user/rst/quickstart.html
.. _pkg_resources: https://setuptools.pypa.io/en/latest/pkg_resources.html
'''

import logging
from datetime import datetime
from collections import defaultdict
from pathlib import Path
import multiprocessing as mp
import pkg_resources as pkg
import numpy as np

from ..fingerprint import fingerprint
from ..path import ensure, sanitize_filename
from ..cosette.task import TaskStatus
from ..cosette.pythontask import PythonTask
from ..gavroche.test import TestResult
from .formatter import Formatter
from .mpl import MplPlot
from .test_report import TestReport, TestReportTask
from .verbosity import Verbosity


LOGGER = logging.getLogger(__name__)


class Rst:
    '''Class to convert :class:`~.TestResult` objects into `reStructuredText`_
    format.'''

    def __init__(self, representation, *, n_workers=None):
        '''Initialize a class instance with the given representation.

        :param Representation representation: A representation.
        '''
        self.representation = representation
        self.formatter = RstFormatter()
        self.plots = {}
        self.tree_dict = defaultdict(list)
        self.text_dict = defaultdict(list)
        self.n_workers = n_workers

    def clear(self):
        '''Clear the content of `self`.'''
        self.plots.clear()
        self.tree_dict.clear()
        self.text_dict.clear()

    def format_report(self, *, report, author, version):
        '''Format a report.

        Most of the work is actually done in the :meth:`format_report_rec`
        method.

        :param TestReport report: the report to format.
        :returns: the formatted report.
        :rtype: FormattedRst
        '''
        self.clear()

        self.format_report_rec(report=report, tree=())

        fmt_report = FormattedRst(author=author, title=report.title,
                                  version=version,
                                  tree_dict=self.tree_dict,
                                  text_dict=self.text_dict,
                                  plots=self.plots,
                                  n_workers=self.n_workers)
        LOGGER.debug('formatted report: %s', fmt_report)
        return fmt_report

    def format_report_rec(self, *, report, tree):
        '''Recursively format report sections and populate the
        ``tree_dict`` and ``text_dict`` attributes.

        Report sections, as represented by the :class:`~.TestReport` class, are
        essentially trees, with :class:`~.TestResult` objects playing the role
        of the tree leaves and :class:`~.TestReport` objects playing the role
        of the tree nodes. This method performs a recursive, pre-order,
        depth-first traversal of the tree and fills the ``tree_dict`` and
        ``text_dict`` dictionaries with the information that is necessary
        to instantiate the final :class:`FormattedRst` object.

        Given a report object, this method constructs a dictionary key by
        concatenating in a tuple the titles of the parent reports and the
        present report. The tuple containing the parent reports' titles is
        recursively passed as the `tree` argument (and it is initialized to the
        empty tuple in the call from :meth:`format_report`). For instance,
        consider the following simplified report structure:

            >>> from valjean.javert.test_report import TestReport
            >>> subsub1 = TestReport(title='SubSub1')
            >>> subsub2 = TestReport(title='SubSub2')
            >>> sub1 = TestReport(title='Sub1', content=[subsub1, subsub2])
            >>> sub2 = TestReport(title='Sub2')
            >>> main = TestReport(title='Main', content=[sub1, sub2],
            ...                   text='Si six scies scient six cyprès, '
            ...                        'six-cent-six scies scient '
            ...                        'six-cent-six cyprès.')

        We instantiate an :class:`Rst` object and we format the report:

            >>> import valjean.javert.representation as rpr
            >>> from valjean.javert.rst import Rst
            >>> rst = Rst(rpr.Representation(rpr.FullRepresenter()))
            >>> formatted_report = rst.format_report(report=main, author='Me',
            ...                                      version='0.1')

        Let us look at the dictionary keys:

            >>> print(sorted(rst.text_dict.keys()))
            [(), ('Sub1',), ('Sub1', 'SubSub1'), ('Sub1', 'SubSub2'), \
('Sub2',)]

        The empty tuple, ``()``, is associated to the main report. The other
        keys represent subreports. The length of the tuple corresponds to the
        depth of the nested report; for instance, the ``'SubSub2'`` section
        appears in the ``('Sub1', 'SubSub2')`` tuple with length 2, because
        ``'SubSub2'`` is nested twice (``Main``/``Sub1``/``SubSub2``).

        The values of the ``text_dict`` dictionary are lists of strings
        representing the text of the given section. For instance, here is what
        the main section looks like:

        .. doctest::

            >>> print('\\n'.join(rst.text_dict[()]))
            Main
            ====
            <BLANKLINE>
            Si six scies scient six cyprès, six-cent-six scies scient \
six-cent-six cyprès.
            <BLANKLINE>

        The ``tree_dict`` dictionary associates a tuple representing a
        report section to the list of tuples that represent the subsections:

            >>> print(rst.tree_dict[()])
            [('Sub1',), ('Sub2',)]
            >>> print(rst.tree_dict[('Sub1',)])
            [('Sub1', 'SubSub1'), ('Sub1', 'SubSub2')]

        If a section does not have any subsection, the corresponding tuple does
        not appear in the dictionary:

            >>> ('Sub2',) in rst.tree_dict
            False

        The ``tree_dict`` and ``text_dict`` dictionaries contain most
        of the information required by :class:`FormattedRst` to write out the
        report in the form of a `reStructuredText`_ file tree (see
        :meth:`~.FormattedRst.write`).
        '''
        LOGGER.debug('formatting tree: %s', tree)
        LOGGER.debug('formatting report: %s', report.title)
        report_text = self.format_section(report, depth=len(tree))
        self.text_dict[tree].extend(report_text)
        for stuff in report.content:
            if isinstance(stuff, TestReport):
                subtree = tree + (stuff.title,)
                self.tree_dict[tree].append(subtree)
                self.format_report_rec(report=stuff, tree=subtree)
            elif isinstance(stuff, TestResult):
                LOGGER.debug('tree: %s', tree)
                LOGGER.debug('tree_dict: %s', dict(self.tree_dict))
                assert tree in self.text_dict
                res_text = self.format_result(stuff)
                self.text_dict[tree].extend(res_text)
            else:
                raise TypeError('expected TestReport or TestResult in '
                                f'TestReport content, found {type(stuff)} '
                                'instead')

    def format_section(self, section, *, depth):
        '''Format a report section.

        :param TestReport report: the report section to format.
        :returns: the formatted report section.
        :rtype: str
        '''
        lines = [self.formatter.header(section.title, depth),
                 self.formatter.text(section.text),
                 '']
        LOGGER.debug('formatted section: %s', lines)
        return lines

    def format_result(self, result):
        '''Format one test result.

        :param TestResult result: A :class:`~.TestResult`.
        :returns: the formatted test result.
        :rtype: str
        '''
        lines = [self.formatter.anchor(fingerprint(result.test)),
                 self.formatter.text(result.test.description), '']
        res_repr = self.representation(result)
        if (not res_repr
                and self.representation.verbosity in (Verbosity.SILENT,
                                                      Verbosity.SUMMARY)):
            return []
        for template in res_repr:
            fmt = self.formatter.template(template)
            lines.append(str(fmt))
            if isinstance(fmt, RstPlot):
                self.plots[fmt.fingerprint] = fmt.mpl_plot
        LOGGER.debug('formatted result: %s', lines)
        return lines


class RstFormatter(Formatter):
    '''Class that dispatches the task of formatting templates as
    `reStructuredText`_. The concrete formatting is handled by separate classes
    (:class:`RstTable`...).
    '''

    HEADER_CHARS = ('=', '-', '`', "'", '"')

    def header(self, name, depth):
        '''Produce the header for formatting a :class:`~.TestResult`.

        :param str name: A header name
        :param int depth: the depth of this header, 0 being the top.
        :returns: the test header, for inclusion in a reST document.
        :rtype: str
        '''
        if depth >= len(self.HEADER_CHARS):
            raise ValueError(f'maximum depth exceeded: {depth} > '
                             f'{self.HEADER_CHARS}')
        lines = [name]
        lines.append(self.HEADER_CHARS[depth]*len(name))
        lines.append('')
        return '\n'.join(lines)

    def text(self, text):
        '''Format some text.

        :param str text: some text to include.
        :returns: the text, in rst format.
        :rtype: str
        '''
        return text

    @staticmethod
    def anchor(fingerprint):
        '''Format an anchor with the given fingerprint.'''
        return f'\n.. _anchor_{fingerprint}:\n\n'

    @staticmethod
    def format_tabletemplate(table):
        '''Format a :class:`~.TableTemplate`.

        :param TableTemplate table: A table.
        :returns: the reST table.
        :rtype: RstTable
        '''
        return RstTable(table)

    @staticmethod
    def format_plottemplate(plot):
        '''Format a :class:`~.PlotTemplate`.

        :param PlotTemplate plot: A plot.
        :returns: the formatted plot
        :rtype: RstPlot
        '''
        return RstPlot(plot)

    @staticmethod
    def format_texttemplate(text):
        '''Format a :class:`~.TextTemplate`.

        :param TextTemplate text: A text with its highlight positions
        :returns: the formatted text (text itself)
        :rtype: str
        '''
        return RstText(text)


class RstTable:
    '''Convert a :class:`~.TableTemplate` into a `reStructuredText`_ table.'''

    HIGHLIGHT_ROLE = 'hl'
    COL_SEP = '  '

    def __init__(self, table, num_fmt='{:11.6g}'):
        '''Construct an :class:`RstTable` from the given
        :class:`~.templates.TableTemplate`.

        :param TableTemplate table: The table to convert.
        :param str num_fmt: A :func:`format` string to specify how numerical
                            table entries should be represented. The default
                            value for this option is ``'{:11.6g}'``.
        '''
        self.table = table
        self.num_fmt = num_fmt

    def __str__(self):
        '''Yield the table, as a reST string. This is probably the method that
        you want to call.'''
        columns = self.table.columns
        highlights = self.table.highlights
        rows = list(self.format_columns(columns, highlights, self.num_fmt))
        table = self.tabularize(self.table.headers, rows, indent=4)
        header = ('.. role:: ' + RstTable.HIGHLIGHT_ROLE + '\n\n'
                  '.. table::\n    :widths: auto\n\n')
        return header + table + '\n'

    @classmethod
    def tabularize(cls, headers, rows, *, indent=0):
        '''Transform a list of headers and a list of rows into a nice reST
        table.

        The `headers` argument must be a list of strings. The `rows` argument
        represents the table rows, as a list of lists of strings. Each sublist
        represents a table row, and it must have the same length as `headers`,
        or terrible things will happen.

        The column widths are automatically computed to accommodate the largest
        template in each column. Smaller templates are automatically
        right-justified.

        .. doctest::

            >>> headers = ['name', 'quest', 'favourite colour']
            >>> rows = [['Lancelot', 'to seek the Holy Grail', 'blue'],
            ...         ['Galahad', 'to seek the Holy Grail', 'yellow']]
            >>> table = RstTable.tabularize(headers, rows)
            >>> print(table)
            ========  ======================  ================
              name            quest           favourite colour
            ========  ======================  ================
            Lancelot  to seek the Holy Grail              blue
             Galahad  to seek the Holy Grail            yellow
            ========  ======================  ================
            <BLANKLINE>

        You can also indent the table by a given amount of spaces with the
        `indent` keyword argument:

        .. doctest::

            >>> table = RstTable.tabularize(headers, rows, indent=4)
            >>> print(table)
                ========  ======================  ================
                  name            quest           favourite colour
                ========  ======================  ================
                Lancelot  to seek the Holy Grail              blue
                 Galahad  to seek the Holy Grail            yellow
                ========  ======================  ================
            <BLANKLINE>

        :param list(str) headers: The table headers.
        :param list(list(str)) rows: The table rows.
        :returns: The reST table, as a string.
        '''
        widths = cls.compute_column_widths(headers, rows)
        LOGGER.debug('widths: %s', widths)

        sep_row = cls.COL_SEP.join('='*w for w in widths)
        header_row = cls.COL_SEP.join(f'{header:^{w}}'
                                      for w, header in zip(widths, headers))
        lines = [' '*indent + sep_row, header_row, sep_row]
        lines.extend(cls.concat_rows(widths, rows))
        lines.append(sep_row)
        lines.append('')  # in order to get 2 tables
        lines_sep = '\n' + ' '*indent
        return lines_sep.join(lines)

    @staticmethod
    def transpose(columns):
        r'''Given a matrix as a list of columns, yield the matrix rows.
        (Equivalently, if the matrix is given as a list of rows, yield the
        columns). For instance, consider the following list:

            >>> matrix = [(11, 21, 31), (12, 22, 32), (13, 23, 33)]
            >>> for column in matrix:
            ...     print(' '.join(str(elem) for elem in column))
            11 21 31
            12 22 32
            13 23 33

        If the tuples are interpreted as columns, `matrix` represents the
        following matrix:

        .. math::

            \begin{pmatrix}11&12&13\\21&22&23\\31&32&33\end{pmatrix}

        Transposing it yields:

            >>> transposed = RstTable.transpose(matrix)
            >>> for row in transposed:
            ...     print(' '.join(str(elem) for elem in row))
            11 12 13
            21 22 23
            31 32 33

        .. note::

            The matrix elements returned by :meth:`transpose` are actually
            0-dimensional :class:`numpy.ndarray` objects. For the purpose of
            their further manipulation this is of little consequence, as they
            transparently support most numerical operations.

        :param columns: An iterable yielding columns
        :type columns: list(list) or list(tuple) or list(numpy.ndarray)
        :raises ValueError: if the columns do not have the same length.
        :returns: a generator for the rows of the transposed matrix.
        '''
        LOGGER.debug('columns: %s, %s', type(columns), len(columns))
        try:
            yield from np.nditer(columns)
        except ValueError as exc:
            raise ValueError('all columns must have the same length') from exc

    @classmethod
    def format_columns(cls, columns, highlights, num_fmt):
        '''Transform a bunch of columns (containing arbitrary data: floats,
        bools, ints, strings...) into an iterable of lists of (optionally
        highlighted) strings representing the table **rows**.

        Example:

            >>> cols = [('European swallow', 'African swallow'),
            ...         (27.7, 35.1),
            ...         ('km/h', 'km/h')]
            >>> highs = [[False, False], [False, True], [False, True]]
            >>> for row in RstTable.format_columns(cols, highs, '{:13.8f}'):
            ...     print(row)
            ['European swallow', '  27.70000000', 'km/h']
            ['African swallow', ':hl:`35.10000000`', ':hl:`km/h`']

        :param columns: An iterable yielding columns. The table columns may
                        contain any kind of data; if the data type is numeric,
                        it will be formatted according to the `num_fmt`
                        argument; other types will just get stringified
                        (:class:`str`).
        :param highlights: An iterable yielding a collection of bools for each
                           column.  The `j`-th element of the `i`-th iteration
                           decides whether the element in the `j`-th column of
                           the `i`-th row should be highlighted.
        :param num_fmt: The format string to be used for numerical types.
        :returns: a generator yielding lists of strings, representing the table
                  rows.
        '''
        def _format_val(val):
            try:
                type_ = val.dtype.type
            except AttributeError:
                return str(val)
            if issubclass(type_, (float, np.inexact)):
                return num_fmt.format(val)
            return str(val)
        for row, hlrow in zip(cls.transpose(columns),
                              cls.transpose(highlights)):
            yield [cls.highlight(_format_val(val), high)
                   for val, high in zip(row, hlrow)]

    @staticmethod
    def compute_column_widths(headers, rows):
        '''Compute the width of the columns that are necessary to accommodate
        all the headers and table rows.

            >>> headers = ['swallow sub-species', 'airspeed', 'laden']
            >>> rows = [['European', '27.7 km/h', 'no'],
            ...         ['European', '22.0 km/h', 'yes'],
            ...         ['African', '35.1 km/h', 'no'],
            ...         ['African', '26.2 km/h', 'yes']]
            >>> RstTable.compute_column_widths(headers, rows)
            [19, 9, 5]
            >>> [len('swallow sub-species'), len('27.7 km/h'), len('laden')]
            [19, 9, 5]

        :param list(str) headers: A list of column headers.
        :param list(list(str)) rows: A list of table rows.
        :returns: a list of integers indicating how wide each columns must be
                  to accommodate all the table elements and the headers.
        :rtype: list(int)
        '''
        col_widths = [len(header) for header in headers]
        for row in rows:
            for i, val in enumerate(row):
                col_widths[i] = max(col_widths[i], len(val))
        return col_widths

    @classmethod
    def highlight(cls, val, flag):
        '''Wrap `val` in highlight reST markers if flag is `True`.

        >>> RstTable.highlight('dis', False)
        'dis'
        >>> RstTable.highlight('DAT', True)
        ':hl:`DAT`'
        '''
        if flag:
            return f':{cls.HIGHLIGHT_ROLE}:`{val.strip()}`'
        return val

    @classmethod
    def concat_rows(cls, widths, rows, just='>'):
        '''Concatenate the given rows and justify them according to the `just`
        parameter (see the :ref:`python:formatspec`). The columns widths (for
        justification) must be provided using the `width` argument.

        :param list(int) widths: The list of columns widths.
        :param list(list(str)) rows: The list of rows; each row must be a list
                                     of strings.
        :param str just: A format character for the justification (usually one
                         of '<', '^', '>').
        :returns: the concatenated, justified rows.
        :rtype: str
        '''
        for row in rows:
            centered = list(f'{val:{just}{width}}'
                            for width, val in zip(widths, row))
            yield cls.COL_SEP.join(centered)


class RstPlot:
    '''This class models a plot in an `reStructuredText`_ document. It converts
    a :class:`~.PlotTemplate` object into an :class:`~.MplPlot`, and it
    provides the ``.. image::`` directive to include in the .rst file.
    '''
    def __init__(self, plot):
        self.fingerprint = fingerprint(plot)
        self.mpl_plot = MplPlot(plot)

    def __str__(self):
        return f'.. image:: /figures/{self.filename()}\n'

    def filename(self):
        '''Make up a(n almost) unique filename for this plot.

        :returns: the filename from fingerprint (png format)
        :rtype: str
        '''
        return f'plot_{self.fingerprint}.png'


class RstText:
    '''Construct an :class:`RstText` from the given
    :class:`~.templates.TextTemplate`.
    '''
    def __init__(self, text):
        self.text = text

    def __str__(self):
        return self.text.text


class FormattedRst:
    '''This class represents a formatted rst document tree, which typically
    consists of an index file and of several sections.
    '''
    def __init__(self, *, author, title, version,
                 tree_dict, text_dict, plots, n_workers=None):
        # pylint: disable=too-many-arguments
        '''Create a :class:`FormattedRst` object. The `author`, `title` and
        `version` arguments are expected to be strings and are
        self-explanatory.

        The `tree_dict` and `text_dict` arguments must be dictionaries.  The
        `tree_dict` dictionary represents the tree structure of the
        `reStructuredText`_ document, and the `text_dict` represent the
        contents of each section. The report sections, which appear as keys in
        both dictionaries, are expected to be tuples of strings, with each
        string representing an additional layer in the document hierarchy. The
        section contents (the values of `text_dict`) are expected to be lists
        of strings to be written to disk.

        Finally, the `plots` argument is a list of the plots referenced by the
        text sections. The plots will be written to disk with filenames of the
        form :samp:`plot_{fingerprint}.png`, where `fingerprint` is the plot
        fingerprint.

        :param str author: the author of this report.
        :param str title: the title of this report.
        :param str version: the version number for this report.
        :param dict tree_dict: dictionary mapping tuples of sections to lists
            of tuples of sections.
        :param dict text_dict: dictionary mapping tuples of sections to lists
            of strings.
        :param plots: list of plots to be written to disk.
        :type plots: list(MplPlot)
        :param n_workers: number of subprocesses to use to write out the plots
            in parallel.  If `None` is given, write the plots in sequential
            mode.
        :type n_workers: int or None
        '''
        if not isinstance(tree_dict, dict):
            raise TypeError("expecting a dictionary as 'tree_dict'"
                            f", got {type(tree_dict)}")
        if not isinstance(text_dict, dict):
            raise TypeError("expecting a dictionary as 'text_dict'"
                            f", got {type(tree_dict)}")

        self.author = author
        self.title = title
        self.version = version
        self.tree_dict = tree_dict.copy()
        self.text_dict = text_dict.copy()
        self.plots = plots.copy()
        self.n_workers = n_workers

    def write(self, path):
        '''Write the text files and the plots into the directory specified by
        `path`.

        :param path: path to the directory that will be written to. It is OK if
                     the directory does not exist.
        :type path: str or pathlib.Path
        '''
        path = Path(path)

        self.setup(path)
        self._write_rec(tree=(), path=path)

        items = [(plot, str(path / 'figures'
                            / f'plot_{fingerprint}.png'))
                 for fingerprint, plot in self.plots.items()]
        if self.n_workers is not None:
            LOGGER.info('writing %d plots using %d subprocesses',
                        len(items), self.n_workers)
            with mp.Pool(self.n_workers) as pool:
                pool.map(self._static_writer, items)
        else:
            LOGGER.info('writing %d plots in sequential mode', len(items))
            for item in items:
                self._static_writer(item)

    @staticmethod
    def _static_writer(plot_path_pair):
        plot_path_pair[0].save(plot_path_pair[1])

    def _write_rec(self, *, tree, path):

        if tree:
            tree_path = self.tree_to_path(base=path, tree=tree)
        else:
            tree_path = path / 'index'

        LOGGER.note('writing tree_path: %s', tree_path)
        ensure(tree_path.parent, is_dir=True)

        subtrees = self.tree_dict[tree]
        write_path = tree_path.with_name(tree_path.name + '.rst')
        with write_path.open('w') as tree_file:
            text = '\n'.join(self.text_dict[tree])
            tree_file.write(text)
            if subtrees:
                toc = self.toc('Contents', subtrees)
                tree_file.write(toc)

        for subtree in subtrees:
            self._write_rec(tree=subtree, path=path)

    @staticmethod
    def tree_to_path(*, base, tree):
        '''Convert a tree to a file path.

        :param base: the base path for all subtrees.
        :type base: pathlib.Path
        :param tuple(str) tree: a sequence of tree nodes, starting from the
                          tree root.
        '''
        paths = [sanitize_filename(node) for node in tree]
        return base.joinpath(*paths)

    def toc(self, toc_title, subtrees):
        '''Build an rst table of contents.

        :param str toc_title: the title for the table of contents (e.g.
                              ``'Contents'``).
        '''
        lines = ['\n\n.. toctree::\n    :titlesonly:\n    '
                 f':caption: {toc_title}:\n']
        for subtree in subtrees:
            # need to remove upper level to subtree in toc else they appear N
            # times in the toc, depending on level, breaking the links
            fname = str(self.tree_to_path(base=Path(''), tree=subtree[-2:]))
            line = ' '*4 + fname
            lines.append(line)
        lines.append('')
        lines.append('')
        return '\n'.join(lines)

    def setup(self, path):
        '''Set up the output directory for :meth:`write`.

        :param pathlib.Path path: the path to the directory.
        '''
        ensure(path, is_dir=True)
        ensure(path / '.static', is_dir=True)
        ensure(path / '.templates', is_dir=True)
        ensure(path / 'figures', is_dir=True)
        year = datetime.now().year
        self.configure(resource='conf.py.template', dest=path / 'conf.py',
                       author=self.author, project=self.title,
                       version=self.version, theme='alabaster',
                       theme_options={}, year=year)
        self.configure(resource='valjean.css',
                       dest=path / '.static' / 'valjean.css',
                       formatting=False)

    @staticmethod
    def configure(resource, dest, formatting=True, **kwargs):
        '''Copy a package resource to the specified destination, optionally
        formatting the resource content using :meth:`str.format`.

        For more information about resources, see `pkg_resources`_.

        :param str resource: the name of the resource.
        :param pathlib.Path dest: the destination path.
        :param bool formatting: whether formatting should be applied.
        :param dict kwargs: any additional keyword arguments will be passed to
                            the formatting.
        '''
        assert pkg.resource_exists('valjean.javert.resources.rst', resource)
        res_template = pkg.resource_string('valjean.javert.resources.rst',
                                           resource).decode('utf-8')
        res_str = res_template.format(**kwargs) if formatting else res_template
        with dest.open('w') as res_file:
            res_file.write(res_str)


class RstTestReportTask(PythonTask):
    '''Task class that transforms a list of tests into a test report.
    :class:`~.TestResult` objects in the environment.'''

    @classmethod
    def from_tasks(cls, name, *, make_report, eval_tasks, representation,
                   author, version, kwargs=None, deps=None, soft_deps=None):
        # pylint: disable=too-many-arguments
        '''Construct an :class:`RstTestReportTask` from a list of test
        evaluation tasks and a function to classify test results and put them
        in test reports.
        '''
        report_name = 'report-' + name
        report_task = TestReportTask(report_name, make_report=make_report,
                                     eval_tasks=eval_tasks, kwargs=kwargs)
        return cls(name, report_task=report_task,
                   representation=representation, author=author,
                   version=version, deps=deps, soft_deps=soft_deps)

    def __init__(self, name, *, report_task, representation, author, version,
                 deps=None, soft_deps=None):
        # pylint: disable=too-many-arguments

        def write_rst(*, env, config):
            report = env[report_task.name]['result']
            if 'args' in config and 'workers' in config['args']:
                n_workers = config['args']['workers']
                LOGGER.debug('will use %d subprocesses to write the report',
                             n_workers)
            else:
                n_workers = None
                LOGGER.debug('will write the report in sequential mode')
            rst = Rst(representation, n_workers=n_workers)
            fmt_report = rst.format_report(report=report, author=author,
                                           version=version)
            report_root = Path(config.query('path', 'report-root'))
            report_path = report_root / sanitize_filename(self.name)
            ensure(report_path, is_dir=True)
            fmt_report.write(report_path)
            env_up = {self.name: {'result': fmt_report,
                                  'output_dir': str(report_path)}}
            return env_up, TaskStatus.DONE

        deps = [] if deps is None else deps
        deps.append(report_task)

        super().__init__(name, write_rst, deps=deps, soft_deps=soft_deps,
                         env_kwarg='env', config_kwarg='config')



# Copyright French Alternative Energies and Atomic Energy Commission
# Contributors: valjean developers
# valjean-support@cea.fr
#
# This software is a computer program whose purpose is to analyze and
# post-process numerical simulation results.
#
# This software is governed by the CeCILL license under French law and abiding
# by the rules of distribution of free software. You can use, modify and/ or
# redistribute the software under the terms of the CeCILL license as circulated
# by CEA, CNRS and INRIA at the following URL: http://www.cecill.info.
#
# As a counterpart to the access to the source code and rights to copy, modify
# and redistribute granted by the license, users are provided only with a
# limited warranty and the software's author, the holder of the economic
# rights, and the successive licensors have only limited liability.
#
# In this respect, the user's attention is drawn to the risks associated with
# loading, using, modifying and/or developing or reproducing the software by
# the user in light of its specific status of free software, that may mean that
# it is complicated to manipulate, and that also therefore means that it is
# reserved for developers and experienced professionals having in-depth
# computer knowledge. Users are therefore encouraged to load and test the
# software's suitability as regards their requirements in conditions enabling
# the security of their systems and/or data to be ensured and, more generally,
# to use and operate it in the same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

'''This module contains somme common utility functions for dynamically
importing Python source files as modules.'''

import sys
import logging
from pathlib import Path


LOGGER = logging.getLogger(__name__)


def split_module_path(file_name):
    '''Split a path to a Python module into a path and a module name.

    :param path: A path to a Python file.
    :type path: str or :term:`path-like object`
    :returns: a `(path, name)` tuple.
    '''
    LOGGER.debug("module_from_name(path='%s')", file_name)
    path = Path(str(file_name)).resolve()
    return str(path.parent), path.stem


def dyn_import(file_name):
    '''Load a Python module from the given file name.

    Note that this function adds the path to :file:`file_name` to ``sys.path``,
    so that local imports in :file:`file_name` work as expected.

    :param str file_name: the name of the file containing the module.
    :returns: the loaded module.
    '''
    from importlib import import_module, invalidate_caches
    LOGGER.debug("dyn_import(file_name='%s')", file_name)
    sfile_name = str(file_name)
    module_path, module_name = split_module_path(sfile_name)
    LOGGER.debug("module_path, name: %s, %s", module_path, module_name)

    # If the module path is not in sys.path, add it.  Ugh, O(n) algorithm here.
    # OTOH, sys.path should never grow terribly large...
    if module_path not in sys.path:
        sys.path.append(module_path)
    else:
        # necessary, otherwise new modules will not be seen
        invalidate_caches()
    module = import_module(module_name)
    return module



# Copyright French Alternative Energies and Atomic Energy Commission
# Contributors: valjean developers
# valjean-support@cea.fr
#
# This software is a computer program whose purpose is to analyze and
# post-process numerical simulation results.
#
# This software is governed by the CeCILL license under French law and abiding
# by the rules of distribution of free software. You can use, modify and/ or
# redistribute the software under the terms of the CeCILL license as circulated
# by CEA, CNRS and INRIA at the following URL: http://www.cecill.info.
#
# As a counterpart to the access to the source code and rights to copy, modify
# and redistribute granted by the license, users are provided only with a
# limited warranty and the software's author, the holder of the economic
# rights, and the successive licensors have only limited liability.
#
# In this respect, the user's attention is drawn to the risks associated with
# loading, using, modifying and/or developing or reproducing the software by
# the user in light of its specific status of free software, that may mean that
# it is complicated to manipulate, and that also therefore means that it is
# reserved for developers and experienced professionals having in-depth
# computer knowledge. Users are therefore encouraged to load and test the
# software's suitability as regards their requirements in conditions enabling
# the security of their systems and/or data to be ensured and, more generally,
# to use and operate it in the same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.
#
# -*- coding: utf-8 -*-
"""This module defines :class:`Env`, a class that makes it simpler to add
information about running tasks. An :class:`Env` object can be created from an
existing dictionary as follows:

    >>> from valjean.cosette.env import Env
    >>> quest = {'name': 'Sir Galahad', 'favourite colour': 'blue'}
    >>> env_quest = Env(quest)  # a shallow copy of `quest` is performed

You can use an :class:`Env` object as a glorified dicitionary (:class:`Env`
inherits from :class:`dict`), but its main purpose is really to store
information about concurrently running tasks (see :class:`~.task.Task`). By
convention, the keys in :class:`Env` are assumed to be task names; the
associated values are dictionaries storing whatever information may be useful
about the task. The dictionaries are also expected to have a ``'status'`` key
describing the current task status (see :class:`~.task.TaskStatus`). An example
of :class:`Env` respecting these conventions is the following:

    >>> from valjean.cosette.task import Task, TaskStatus
    >>> class FindHolyGrail(Task):
    ...   '''We derive a class from Task, which is abstract.'''
    ...   def do(self, env, config):
    ...     # find the Holy Grail
    ...     pass
    >>> quest = FindHolyGrail('quest')  # 'quest' is the task name
    >>> tasks = {quest.name: {'name': 'Sir Galahad',
    ...                       'favourite colour': 'blue',
    ...                       'status': TaskStatus.FAILED}}
    >>> env = Env(tasks)

The :class:`Env` API integrates well with the :mod:`~.task` module and provides
a number of practical methods for dealing with tasks. For instance, there are
``is_*`` methods for all members of the :class:`~.TaskStatus` enumeration:

    >>> env.is_failed(quest)
    True
    >>> print(env.get_status(quest))    # equivalently
    TaskStatus.FAILED

Additionally, you can change the status of a task with

    >>> env.set_status(quest, TaskStatus.DONE)
    >>> env.set_done(quest)  # equivalent, shorter version

Information about the tasks, incuding their status, is stored with the task
name as the key:

    >>> print(env[quest.name]['status'])
    TaskStatus.DONE

The :class:`Env` class tries hard to be thread-safe; that is, all its methods
will operate atomically. Internally, thread safety is enforced by locking the
object whenever its contents are accessed. This, however, does not help in case
of read-and-modify operations, as in the following example::

    >>> if env.is_done(quest):      # WARNING: do not try this at home
    ...     env.set_skipped(quest)  # race condition here!

This snippet is racy in multithreaded mode because another thread may change
the status of ``quest`` after :meth:`~.is_done()` has released the lock but
before :meth:`~.set_skipped()` has had the chance to acquire it. For these
scenarios, :class:`Env` offers the :meth:`~.atomically()` method, which accepts
as an argument the action that should be performed.  When called,
:meth:`~.atomically()` first acquires the lock on the object, and then passes
the :class:`Env` object to the action. A thread-safe implementation of the
read-and-modify trip above is implemented as follows:

    >>> def modify_task1(env):
    ...     if env.is_done(quest):
    ...         env.set_skipped(quest)
    >>> env.atomically(modify_task1)
    >>> env.is_skipped(quest)
    True
"""

import logging
import threading
import pickle
from collections.abc import MutableMapping

from .task import TaskStatus


LOGGER = logging.getLogger(__name__)


class EnvError(Exception):
    '''An error that may be raised by the :class:`~Env` class.'''


def _add_enum_accessors(enum):
    '''Dynamically instantiate is_* and set_* methods for the :class:`Env`
    class, based on the values of the :class:`~.TaskStatus` enum.
    '''
    def _decorator(cls):
        def _getter_factory(name, status):
            '''Produce a getter with a suitable docstring.'''
            def _getter(self, task):
                return self.get_status(task) == status
            _getter.__doc__ = f"Returns True if `task`'s status is `{name}`."
            return _getter

        def _setter_factory(name, status):
            def _setter(self, task):
                self.set_status(task, status)
            _setter.__doc__ = f"Sets `task`'s status to `{name}`."
            return _setter

        for name, status in enum.__members__.items():
            lname = name.lower()

            # add the getter
            getter = _getter_factory(name, status)
            getter_name = 'is_' + lname
            setattr(cls, getter_name, getter)

            # add the setter
            setter = _setter_factory(name, status)
            setter_name = 'set_' + lname
            setattr(cls, setter_name, setter)
        return cls
    return _decorator


@_add_enum_accessors(TaskStatus)
class Env(MutableMapping):
    '''The :class:`Env` class can be used to store and dynamically update
    information about concurrently running tasks, and offers thread-safety
    guarantees.
    '''

    def __init__(self, dictionary=None):
        '''Construct an object based on an existing dictionary.'''
        self.dictionary = {}
        self.lock = threading.RLock()
        if dictionary is not None:
            with self.lock:
                self.update(dictionary)

    def __getitem__(self, key):
        return self.dictionary[key]

    def __setitem__(self, key, value):
        self.dictionary[key] = value

    def __delitem__(self, key):
        del self.dictionary[key]

    def __iter__(self):
        yield from self.dictionary

    def __len__(self):
        return len(self.dictionary)

    def __repr__(self):
        return f'{self.__class__.__name__}({self.dictionary!r})'

    @classmethod
    def from_file(cls, path, *, fmt='pickle'):
        '''Deserialize an :class:`Env` object from a file.

        :param str path: Path to the file.
        :param str fmt: Serialization format (only ``'pickle'`` is supported
            for the moment).
        :returns: The deserialized object.
        '''
        def deserializer(file_):
            return pickle.load(file_)
        mode = 'rb'
        try:
            with open(path, mode) as input_file:
                deser = deserializer(input_file)
        except IOError as error:
            if error.errno == 2:
                LOGGER.debug('environment file %r is missing, returning an '
                             'empty environment', path)
            else:
                LOGGER.warning('cannot load %s environment from file %r. '
                               'Error message: %s', fmt, path, error.strerror)
            return None
        except ValueError as error:
            LOGGER.warning('cannot load %s environment from file %r. '
                           'Error message: %s', fmt, path, error)
            return None
        LOGGER.debug('returning environment: %s', deser)
        return deser

    def to_file(self, path, *, task_name=None, fmt='pickle'):
        '''Serialize an :class:`Env` object to a file.

        :param str path: path to the file.
        :param task_name: name of the task to serialize, or `None` to serialize
            the whole environment.
        :type task_name: str or None
        :param str fmt: serialization format (only ``'pickle'`` is supported
            for the moment).
        '''
        def serializer(file_):
            if task_name is None:
                pickle.dump(self, file_)
                return
            partial_env = Env({task_name: self[task_name]})
            pickle.dump(partial_env, file_)
        mode = 'wb'
        try:
            with open(path, mode) as output_file:
                serializer(output_file)
        except IOError as error:
            LOGGER.error("cannot write %s environment to file '%s'.\n"
                         "Error message: %s", fmt, path, error.strerror)

    def merge_done_tasks(self, other):
        '''Merge task status from another environment.

        This method takes an additional environment `other` as an argument. If
        the same key appears in `self` and `other` and ``other[key]['status']
        == TaskStatus.DONE``, then it sets ``self[key] = other[key]``.

        The idea is that `self` might contain a pristine environment, while
        `other` might provide the results of a previous run. We want to mark
        completed tasks as `DONE`, but we also want to re-run those that
        failed.

        :param Env other: The environment providing the updates.
        '''

        # initialize the environment
        for task_name, status in other.items():
            if status['status'] != TaskStatus.DONE:
                continue
            LOGGER.debug('merging status for task %s', task_name)
            self[task_name] = status

    def set_status(self, task, status):
        '''Set `task`'s status to `status`.'''
        with self.lock:
            self.setdefault(task.name, {})['status'] = status

    def get_status(self, task):
        '''Return `task`'s status.'''
        with self.lock:
            status = self.setdefault(task.name,
                                     {'status': TaskStatus.WAITING})['status']
            return TaskStatus(status)

    def atomically(self, action):
        '''Perform an action atomically on the environment dictionary. The
        dictionary passes itself as the first argument to `action`, which must
        be callable.
        '''
        with self.lock:
            return action(self)

    def apply(self, env_update):
        '''Apply un update to the dictionary.'''

        if env_update is None:
            return

        def _apply_worker(update, old):
            for key, val in update.items():
                if isinstance(val, MutableMapping):
                    if key in old:
                        # recursively update this dictionary
                        _apply_worker(val, old[key])
                    else:
                        old[key] = val
                else:
                    old[key] = val

        with self.lock:
            _apply_worker(env_update, self)

    def set_start_end_clock(self, task, *, start, end):
        '''Set the start and end time for the given task.'''
        env_update = {task.name: {'start_clock': start,
                                  'end_clock': end}}
        self.apply(env_update)

    def get_start_clock(self, task):
        '''Return the start time for the given task.'''
        return self.dictionary.get(task.name, None).get('start_clock', None)

    def get_end_clock(self, task):
        '''Return the end time for the given task.'''
        return self.dictionary.get(task.name, None).get('end_clock', None)

    def copy(self):
        '''Return a shallow copy of `self`.'''
        return Env(self.dictionary.copy())

    def __getstate__(self):
        '''Do not serialize the lock, as doing so results in exceptions with
        Python >=3.6.'''
        state = self.__dict__.copy()
        del state['lock']
        return state

    def __setstate__(self, state):
        '''The serialized state does not contain a lock; create a new one
        instead.'''
        self.__dict__.update(state)
        self.lock = threading.RLock()



# Copyright French Alternative Energies and Atomic Energy Commission
# Contributors: valjean developers
# valjean-support@cea.fr
#
# This software is a computer program whose purpose is to analyze and
# post-process numerical simulation results.
#
# This software is governed by the CeCILL license under French law and abiding
# by the rules of distribution of free software. You can use, modify and/ or
# redistribute the software under the terms of the CeCILL license as circulated
# by CEA, CNRS and INRIA at the following URL: http://www.cecill.info.
#
# As a counterpart to the access to the source code and rights to copy, modify
# and redistribute granted by the license, users are provided only with a
# limited warranty and the software's author, the holder of the economic
# rights, and the successive licensors have only limited liability.
#
# In this respect, the user's attention is drawn to the risks associated with
# loading, using, modifying and/or developing or reproducing the software by
# the user in light of its specific status of free software, that may mean that
# it is complicated to manipulate, and that also therefore means that it is
# reserved for developers and experienced professionals having in-depth
# computer knowledge. Users are therefore encouraged to load and test the
# software's suitability as regards their requirements in conditions enabling
# the security of their systems and/or data to be ensured and, more generally,
# to use and operate it in the same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

'''External test used to represent user formatted results in the report.

This test has been designed in order to offer the possibility to define and run
user-defined tests outside valjean and include their results in a report, along
with the native valjean tests.

The user is responsible for defining and representing the test results. The
representation must be done with templates (see :mod:`.templates` module).

For example, the simplest case is a custom plot to represent a test. In that
case, if the path to the image is known a :class:`~.templates.TextTemplate` can
be created and given to the test:

>>> txtemp = TextTemplate("And this represents Javert\\n\\n"
...                       ".. image:: ../../doc/src/images/javert.jpg\\n\\n")
>>> textr = TestExternal(txtemp, name='', success=True).evaluate()
>>> print(bool(textr))
True
'''

from ..gavroche.test import Test, TestResult
from .templates import TextTemplate, PlotTemplate, TableTemplate


class TestResultExternal(TestResult):
    '''Result from a external test.'''

    def __bool__(self):
        '''This test returns the return value specified in the test
        initialisation.
        '''
        return self.test.success


class TestExternal(Test):
    '''Class building a test that will return the boolean value specified at
    initialisation.
    '''

    def __init__(self, *templates, name, description='', labels=None,
                 success=True):
        '''Initialize the :class:`~.TestExternal` object with the templates to
        represent in the report, a name, a description of the test (may be
        long).

        External test are performed by the user who only gives its
        representation as templates and its return value. The return value can
        be an arbitrary value (a test that always returns `True` or `False`)
        for example.

        :param templates: templates representing the external test
        :type templates: TextTemplate, PlotTemplate, TableTemplate
        :param str name: name of the test
        :param str description: description of the test expected with context
        :param dict labels: labels for test classification
        :param bool success: returned value by :class:`TestResultExternal`
        '''
        super().__init__(name=name, description=description, labels=labels)
        if any(not isinstance(t, (TextTemplate, PlotTemplate, TableTemplate))
               for t in templates):
            raise TypeError('Expected templates')
        self.templates = templates
        self.success = success

    def evaluate(self):
        '''Evaluate the test (do nothing in this case except building the
        TestResult).

        :rtype: TestResultExternal
        '''
        return TestResultExternal(self)



# Copyright French Alternative Energies and Atomic Energy Commission
# Contributors: valjean developers
# valjean-support@cea.fr
#
# This software is a computer program whose purpose is to analyze and
# post-process numerical simulation results.
#
# This software is governed by the CeCILL license under French law and abiding
# by the rules of distribution of free software. You can use, modify and/ or
# redistribute the software under the terms of the CeCILL license as circulated
# by CEA, CNRS and INRIA at the following URL: http://www.cecill.info.
#
# As a counterpart to the access to the source code and rights to copy, modify
# and redistribute granted by the license, users are provided only with a
# limited warranty and the software's author, the holder of the economic
# rights, and the successive licensors have only limited liability.
#
# In this respect, the user's attention is drawn to the risks associated with
# loading, using, modifying and/or developing or reproducing the software by
# the user in light of its specific status of free software, that may mean that
# it is complicated to manipulate, and that also therefore means that it is
# reserved for developers and experienced professionals having in-depth
# computer knowledge. Users are therefore encouraged to load and test the
# software's suitability as regards their requirements in conditions enabling
# the security of their systems and/or data to be ensured and, more generally,
# to use and operate it in the same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

r'''
.. _scipy: https://docs.scipy.org/doc/scipy/reference/index.html

In Student's t-test, we consider 2 different datasets that can have different
means and different variances. The main problem is to test if the means are
equal.  One way to do that is to use Student's t-test, although technically
this name should be reserved for the case where the variances are unknown but
they are known to be equal (the generalized version to different variances is
called Welch's test).

In Tripoli-4, like in most Monte Carlo codes, the estimated statistical
uncertainty is the standard error, i.e. the standard deviation of the mean of
the samples. It is typically assumed that the distribution of the mean is
normal, as a consequence of the central limit theorem. The variance of the
population is estimated as

.. math::

    s^2 = \frac{1}{n-1}\sum_{i=1}^n(x_i - \bar{x})^2

and the variance of the mean is estimated as

.. math::

    \bar{s}^2 = \frac{s^2}{n}

The standard error :math:`\bar{s}` is the quantity that typically enters
Student's t-test.

The Student test statistic is:

.. math::

    t_{\text{Student}} = \frac{\bar{x}_1 - \bar{x}_2}
    {\sqrt{\bar{s}_1^2 + \bar{s}_2^2}}


Note that, with our :class:`~valjean.eponine.dataset.Dataset`, this is
equivalent to

.. code::

    diff = dataset_1 - dataset_2
    t(Student) = diff.value / diff.error

as value is difference of values and error is equal to the quadratic sum of the
errors.


Test interpretation
```````````````````

The tested hypothesis in the Student's t-test, i.e. the null hypothesis is:

* equal means, so :math:`m_1 = m_2`
* Gaussian errors (or error following a normal distribution)
* independent datasets.

The test interpretation is based on the comparison of the p-value with a given
significance level, called :math:`\alpha`. The (two-sided) p-value is defined
as the probability to observe results more extreme than those that were
actually observed, assuming that the null hypothesis is true. Mathematically,
:math:`p=P(|X| \geq |t_{\text{Student}}|\,|N)`, :math:`X` being the Student
test statistic for samples of size :math:`N`; the variable :math:`N` is also
called the number of degrees of freedom.  The significativity level
:math:`\alpha` is given by the user, usually 1% or 5%.  Student's test fails if
:math:`P(|X| \geq |t_{\text{Student}}|) \leq \alpha`. If this is the case the
null hypothesis is rejected.

To evaluate the test we thus need an estimate of the number of degrees of
freedom, which is typically approximated as :math:`n_1 + n_2 - 2`, with
:math:`n_1` and :math:`n_2` being the number of batches. For large values of
:math:`n_i` (i.e. for a large number of degrees of freedom), the distribution
of the Student statistic approaches a standard normal distribution.

It is sometimes easier to evaluate if the test succeeds by directly comparing
the value of :math:`t_{\text{Student}}` with a threshold given by the
significativity level :math:`\alpha` and the number of degrees of freedom.

For example, here are some thresholds for some values of :math:`\alpha` and of
the number of degrees of freedom:

.. |inf| replace:: :math:`\infty`

+------------+----------------------+----------------------+
|            | :math:`\alpha` = 1 % | :math:`\alpha` = 5 % |
+------------+----------------------+----------------------+
|  N = 10    |        2.7638        |        1.8125        |
+------------+----------------------+----------------------+
|  N = 1000  |        2.5807        |        1.9623        |
+------------+----------------------+----------------------+
|  N = |inf| |        2.5758        |        1.9600        |
+------------+----------------------+----------------------+

Results for N = |inf| come from the normal distribution.


.. note::

    This test can also be used on histograms (spectrum, mesh). In this case, an
    independent test is made for each bin. The test for the whole histogram
    succeeds if *all* the individual tests succeed.  A better option is to use
    a Bonferroni or a Holm-Bonferroni test (see the :mod:`~.bonferroni`
    module).

Code implementation
```````````````````

If the number of degrees of freedom is given, p-values are estimated using
Student's t-distribution. If the number is not given or it is `None`, it is
assumed that it is very large and the normal approximation is used.

The p-value is only calculated if the number of degrees of freedom is given;
otherwise, the Student result is just compared to the threshold. To test the
p-value instead of the value of the t-statistic, use
:func:`TestResultStudent.test_pvalue`.

All these methods use the **SciPy** package (see `scipy`_). The Student
distribution comes from :obj:`scipy.stats.t`, where the pdf is explicitely
written.

The method ``sf`` of :obj:`scipy.stats.t` corresponds to :math:`1 - cdf`, so
this gives the one-sided p-value; we need to multiply it by 2.0 to get the
two-sided p-value. The ``ppf`` (percent point) function returns the threshold
corresponding to the required one-sided significativity, so we need to divide
:math:`\alpha` by 2.0 in our case.

The methods where the test is really applied are static methods, they can be
used outside of the :class:`TestStudent` class. Examples are given in the class
docstrings.


Examples
````````

Let's get an example: one simulation gave 5.3 ± 0.2, the other one 5.25 ± 0.08.
Are these numbers in agreement ?

>>> from valjean.eponine.dataset import Dataset
>>> from valjean.gavroche.stat_tests.student import TestStudent
>>> import numpy as np
>>> ds1 = Dataset(5.3, 0.2)
>>> ds2 = Dataset(5.25, 0.08)
>>> tstudent = TestStudent(ds1, ds2, name="comp",
...                        description="Comparison using Student's t-test")
>>> tstudent_res = tstudent.evaluate()
>>> print(f'{tstudent_res.tstud[0]:.7f}')
0.2321192
>>> print(f'{tstudent_res.test.threshold:.7f}')
2.5758293
>>> bool(tstudent_res)
True

To obtain the p-value, a number of degrees of freedom should be given:

>>> tstudent = TestStudent(ds1, ds2, ndf=1000, name="comp",
...                        description="Comparison using Student's t- test")
>>> tstudent_res = tstudent.evaluate()
>>> print(f'{tstudent_res.tstud[0]:.7f}')
0.2321192
>>> print(f'{tstudent_res.test.threshold:.7f}')
2.5807547
>>> bool(tstudent_res)
True
>>> print(f'{tstudent_res.pvalue[0]:.7f}')
0.8164929
>>> tstudent_res.test_pvalue()
[True]


This is also possible if we want to compare a bin to bin different spectra.
Let's define 2 small datasets containing arrays:

>>> ds3 = Dataset(np.array([5.2, 5.3, 5.25, 5.4, 5.5]),
...               np.array([0.2, 0.25, 0.1, 0.2, 0.3]))
>>> ds4 = Dataset(np.array([5.1, 5.6, 5.2, 5.3, 5.2]),
...               np.array([0.1, 0.3, 0.05, 0.4, 0.3]))
>>> tstudent = TestStudent(ds3, ds4, name="comp",
...                        description="Comparison using Student's t-test")
>>> tstudent_res = tstudent.evaluate()
>>> print(np.array2string(tstudent_res.oracles()))
[[ True  True  True  True  True]]
>>> bool(tstudent_res)
True

In the case of multiple comparisons, the test will return `True` if *all* the
individual comparisons succeed. This is the case here.

It is possible to get a detailed bin-by-bin comparisons, including bad bins:

>>> ds5 = Dataset(np.array([5.1, 5.9, 5.8, 5.3, 4.5]),
...               np.array([0.1, 0.1, 0.05, 0.4, 0.1]))
>>> tstudent = TestStudent(ds3, ds5, name="comp",
...                        description="Comparison using Student's t-test")
>>> tstudent_res = tstudent.evaluate()
>>> print(f'{tstudent_res.test.threshold:.7f}')
2.5758293
>>> print(np.array2string(tstudent_res.oracles()))
[[ True  True False  True False]]

Since some of the comparisons failed here, the whole tests fails as well:

>>> bool(tstudent_res)
False

As in the scalar case, it is possible to ask for the p-values. In both cases
the significativity can be changed via the ``alpha`` argument, as shown here:

>>> tstudent = TestStudent(ds3, ds5, name="comp", alpha=0.05, ndf=1000,
...                        description="Comparison using Student's t-test")
>>> tstudent_res = tstudent.evaluate()
>>> print(f'{tstudent_res.test.threshold:.7f}')
1.9623391
>>> print(np.array2string(tstudent_res.oracles()))
[[ True False False  True False]]
>>> print(f'{tstudent_res.tstud[0][1]:.7f}')
-2.2283441

This last value is in-between the thresholds at 1% and 5%, in magnitude, and
therefore the test succeeds in the former case and fails in the latter.

Obtaining p-values and comparing them is also possible in the spectrum case:

>>> print(np.array2string(tstudent_res.pvalue[0],
...                       formatter={'float_kind':'{:.7e}'.format}))
[6.5481769e-01 2.6079281e-02 1.0147751e-06 8.2310894e-01 1.6125524e-03]
>>> print(np.array2string(tstudent_res.test_pvalue()[0]))
[ True False False  True False]

Finally, for an array of size 1 (and dimension 1), the test will do the same as
for datasets containing :obj:`numpy.generic`:

>>> ds6 = Dataset(np.array([5.3]), np.array([0.2]))
>>> ds7 = Dataset(np.array([5.25]), np.array([0.08]))
>>> tstudent = TestStudent(ds6, ds7, name="comp",
...                        description="Comparison using Student's t-test")
>>> tstudent_res = tstudent.evaluate()
>>> print(np.array2string(tstudent_res.tstud[0],
...                       formatter={'float_kind':'{:.7f}'.format}))
[0.2321192]
>>> print(f'{tstudent_res.test.threshold:.7f}')
2.5758293
>>> bool(tstudent_res)
True

Like for usual operations on datasets, only tests between datasets of the same
shape are possible:

>>> tstudent = TestStudent(ds1, ds7, name="comp",
...                        description="Comparison using Student's t-test")
>>> tstudent_res = tstudent.evaluate()
Traceback (most recent call last):
    ...
ValueError: Datasets to subtract do not have same shape

Multi-dimensional datasets are also allowed:

>>> ds8 = Dataset(np.array([[5.2, 5.3, 5.25], [5.4, 5.5, 5.2]]),
...               np.array([[0.2, 0.25, 0.1], [0.2, 0.3, 0.1]]))
>>> ds9 = Dataset(np.array([[5.1, 5.6, 5.2], [5.3, 5.2, 5.3]]),
...               np.array([[0.1, 0.3, 0.05], [0.4, 0.3, 0.2]]))
>>> ds8.shape
(2, 3)
>>> ds8.size
6
>>> tstudent = TestStudent(ds8, ds9, alpha=0.05, ndf=1000, name="comp",
...                        description="Comparison using Student's t-test")
>>> tstudent_res = tstudent.evaluate()
>>> bool(tstudent_res)
True
>>> print(np.array2string(tstudent_res.tstud[0],
...                       formatter={'float_kind':'{:.7f}'.format}))
[[0.4472136 -0.7682213 0.4472136]
 [0.2236068 0.7071068 -0.4472136]]
>>> print(np.array2string(tstudent_res.oracles()))
[[[ True  True  True]
  [ True  True  True]]]


.. warning::
    If the errors are equal to 0, so the Student denominator is 0, AND the
    values are equal, so the numerator is also 0, the Student value is set to
    0. In that case the test is assumed to succeed (boolean returning `True`).

    >>> ds10 = Dataset(np.array([3.2, 0, 5]), np.array([0, 0.5, 0]))
    >>> ds11 = Dataset(np.array([3.2, 0, 5.2]), np.array([0, 0.5, 0]))
    >>> tstudent = TestStudent(ds10, ds11, name='zeros')
    >>> tstudent_res = tstudent.evaluate()
    >>> bool(tstudent_res)
    False
    >>> print(np.array2string(tstudent_res.tstud[0]))
    [  0.   0. -inf]
    >>> print(np.array2string(tstudent_res.oracles()))
    [[ True  True False]]
'''
import logging
import numpy as np
from scipy.stats import t, norm
from ..test import check_bins, TestDataset, TestResult


LOGGER = logging.getLogger(__name__)


class TestResultStudent(TestResult):
    '''Result from Student's t-test.'''

    def __init__(self, test, tstud, pvalue=None):
        '''Initialisation of :class:`~.TestResultStudent`

        Members are lists whose length corresponds to the number of datasets
        compared to the reference dataset.

        :param test: the TestStudent object
        :type test: :class:`~.TestStudent`
        :param tstud: Student's t-test results
        :type tstud: :class:`list` (:obj:`numpy.ndarray`)
        :param pvalue: p-value or p-values depending on datasets, default is
                       None
        :type pvalue: :class:`list` (:class:`numpy.generic`)
            or :class:`list` (:class:`list` (:class:`numpy.generic`))
        '''
        super().__init__(test)
        self.tstud = tstud
        self.pvalue = pvalue

    def test_alpha(self, tstud):
        '''Test p-value or first kind error.

        :param tstud: value of Student's t-statistic to be used for comparison
        :type tstud: :obj:`numpy.generic`
        :returns: bool
        '''
        return np.less(np.fabs(tstud), self.test.threshold)

    def oracles(self):
        '''Final test (if spectrum)

        :returns: list(bool)
        '''
        if isinstance(self.tstud, np.generic):
            return np.array([self.test_alpha(self.tstud)])
        return self.test_alpha(self.tstud)

    def __bool__(self):
        '''Has this test succeeded?

        The policy for :class:`TestResultStudent` is that the test is
        considered to fail if any of the p-values are below the given
        significance level. This means that a test on a :class:`~.Dataset`
        having multiple bins will fail if *any* of the bins exhibits large
        fluctuations. You might want to use :class:`~.TestBonferroni` or
        :class:`~.TestHolmBonferroni` to deal with such cases.

        If this test concerns multiple datasets, the test is considered to
        succeed if all the dataset individually pass the test against the
        reference.
        '''
        result = True
        for tstud in self.tstud:
            if isinstance(tstud, np.generic):
                result = result and bool(self.test_alpha(tstud))
            else:
                result = result and bool(self.test_alpha(tstud).all())
        return result

    def test_pvalue(self):
        '''Result of the test by testing p-value.

        :returns: bool or :obj:`numpy.ndarray` of bool if ``ndf`` was given to
                 :class:`TestStudent` and datasets are :obj:`numpy.ndarray`
        '''
        if self.test.ndf is None:
            return False
        return [pval > self.test.alpha for pval in self.pvalue]


class TestStudent(TestDataset):
    '''Class to build the Student's t-test.'''

    def __init__(self, dsref, *datasets, name, description='', labels=None,
                 alpha=0.01, ndf=None):
        # pylint: disable=too-many-arguments
        '''Initialisation of :class:`TestStudent`

        :param str name: local name of the test
        :param str description: specific description of the test
        :param dict labels: labels to be used for test classification in
                            reports (for example category, input file name,
                            type of result, ...)
        :param dsref: reference dataset
        :type dsref: :class:`~valjean.eponine.dataset.Dataset`
        :param datasets: list of datasets to be compared to reference dataset
        :type datasets: :class:`list` (:class:`~.dataset.Dataset`)
        :param float alpha: significance level, i.e. limit on the p-value
                            (expected values are typically 0.01, 0.05),
                            default is ``0.01``
        :param int ndf: default is ``None``, if given p-value will be
                        calculated for ndf degrees of freedom
                        (should correspond to number of batches),
                        otherwise ndf assumed infinite, normal approximation
        '''
        super().__init__(dsref, *datasets,
                         name=name, description=description, labels=labels)
        self.alpha = alpha
        self.ndf = ndf
        self.threshold = self.student_threshold(self.alpha, self.ndf)

    def evaluate(self):
        '''Evaluate Student's t-test method.

        :returns: :class:`~.TestResultStudent`
        '''
        tstuds = []
        for _ds in self.datasets:
            check_bins(self.dsref, _ds)
            tstuds.append(self.student_test(self.dsref, _ds))
        pval = [self.pvalue(tstud, self.ndf) for tstud in tstuds]
        return TestResultStudent(self, tstuds, pval)

    @staticmethod
    def student_test(ds1, ds2):
        '''Compute Student's t-test or Student distribution for the given
        datasets (**static method**).

        :param ds1: dataset 1
        :type ds1: :class:`~valjean.eponine.dataset.Dataset`
        :param ds2: dataset 2
        :type ds2: :class:`~valjean.eponine.dataset.Dataset`
        :returns: :obj:`numpy.generic` or :obj:`numpy.ndarray` depending on
                  :class:`~valjean.eponine.dataset.Dataset` type
        '''
        diff = ds1 - ds2
        studentt = diff.value / diff.error
        if isinstance(diff.error, np.generic):
            if diff.error == 0 and diff.value == 0:
                LOGGER.debug('Student test set to 0 as giving 0 / 0')
                return np.zeros_like(ds1.value)
            if np.isnan(ds1.error) and np.isnan(ds2.error) and diff.value == 0:
                LOGGER.debug('Student test set to 0 as giving 0 / nan with '
                             'nan coming from both datasets.')
                return np.zeros_like(ds1.value)
            if np.isnan(ds1.value) and np.isnan(ds2.value):
                LOGGER.debug('Student test set to 0 as both values are nan')
                return np.zeros_like(ds1.value)
            return studentt
        studentt[(diff.value == 0) & (diff.error == 0)] = 0
        studentt[(diff.value == 0) & np.isnan(ds1.error)
                 & np.isnan(ds2.error)] = 0
        studentt[np.isnan(ds1.value) & np.isnan(ds2.value)] = 0
        return studentt

    @staticmethod
    def pvalue(tstud, ndf):
        '''Calculation of the p-value (**static method**).

        :param tstud: Student's t-test statistic
        :type tstud: :obj:`numpy.generic` (:obj:`float`) or
                     :obj:`numpy.ndarray` (:obj:`float`)
        :param ndf: number of degrees of freedom
        :type ndf: int or :obj:`numpy.generic` (:obj:`int`)
        :returns: :obj:`numpy.generic` (:obj:`float`) or
                  :obj:`numpy.ndarray` (:obj:`float`)
        '''
        if ndf is None:
            return 2.0*norm.sf(np.fabs(tstud))
        return 2.0*t.sf(np.fabs(tstud), ndf)

    @staticmethod
    def student_threshold(alpha, ndf=None):
        r'''Get threshold from probability (**static method**).

        :param alpha: required significance level (:math:`\alpha`)
        :type alpha: float or :obj:`numpy.generic` (:obj:`float`)
        :param ndf: number of degrees of freedom
        :type ndf: int or :obj:`numpy.generic` (:obj:`int`), default is
                   ``None``
        :returns: :obj:`numpy.generic` (:obj:`float`)

        The threshold is two-sided.

        If number of degrees of freedom is given, the distribution used is the
        Student distribution corresponding to ``ndf``.  If not, we consider an
        infinite number of degrees of freedom.  In that case, the Student
        distribution tends to a normal distribution.
        '''
        if ndf is None:
            return np.fabs(norm.ppf(0.5*alpha))
        return np.fabs(t.ppf(0.5*alpha, ndf))

    def data(self):
        '''Generator yielding objects supporting the buffer protocol that (as a
        whole) represent a serialized version of `self`.'''
        yield from super().data()
        yield self.__class__.__name__.encode('utf-8')
        yield float(self.alpha).hex().encode('utf-8')
        if self.ndf is not None:
            yield hex(self.ndf).encode('utf-8')



# Copyright French Alternative Energies and Atomic Energy Commission
# Contributors: valjean developers
# valjean-support@cea.fr
#
# This software is a computer program whose purpose is to analyze and
# post-process numerical simulation results.
#
# This software is governed by the CeCILL license under French law and abiding
# by the rules of distribution of free software. You can use, modify and/ or
# redistribute the software under the terms of the CeCILL license as circulated
# by CEA, CNRS and INRIA at the following URL: http://www.cecill.info.
#
# As a counterpart to the access to the source code and rights to copy, modify
# and redistribute granted by the license, users are provided only with a
# limited warranty and the software's author, the holder of the economic
# rights, and the successive licensors have only limited liability.
#
# In this respect, the user's attention is drawn to the risks associated with
# loading, using, modifying and/or developing or reproducing the software by
# the user in light of its specific status of free software, that may mean that
# it is complicated to manipulate, and that also therefore means that it is
# reserved for developers and experienced professionals having in-depth
# computer knowledge. Users are therefore encouraged to load and test the
# software's suitability as regards their requirements in conditions enabling
# the security of their systems and/or data to be ensured and, more generally,
# to use and operate it in the same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

'''This module define tests for metadata.'''

from ..test import Test, TestResult


class Missing:
    '''Class for missing metadata.'''

    def __str__(self):
        return 'MISSING'


MISSING = Missing()


class TestResultMetadata(TestResult):
    '''Results of metadata comparisons.'''

    def __init__(self, test, dict_res):
        '''Initialisation of TestResultMetadata.'''
        super().__init__(test)
        self.dict_res = dict_res

    def __bool__(self):
        # aitem = [all(val.values()) for val in self.dict_res.values()]
        # return all(aitem)
        return all(v for val in self.dict_res.values() for v in val.values())

    def per_key(self):
        '''Test result sorted by key.'''
        ditem = {key: all(val.values()) for key, val in self.dict_res.items()}
        return ditem

    def only_failed_comparisons(self):
        '''Return only the failed comparisons. Structure is the same as the
        ``dict_res``.'''
        ditem = {key: self.test.all_md[key]
                 for key, val in self.dict_res.items()
                 if not all(val.values())}
        return ditem


class TestMetadata(Test):
    '''A test that compares metadata.

    .. todo::

        Document the parameters...
    '''

    def __init__(self, dict_md, name, description='', labels=None,
                 exclude=('results', 'index', 'score_index', 'response_index',
                          'response_type')):
        '''Initialisation of :class:`TestMetadata`.

        :param str name: local name of the test
        :param str description: specific description of the test
        :param dict labels: labels to be used for test classification in
            reports, for example category, input file name, type of result, ...
        :param tuple exclude: a tuple of keys that will not be considered as
            metadata. Default: ``('results', 'index', 'score_index',
            'response_index', 'response_type')``
        '''
        super().__init__(name=name, description=description, labels=labels)
        if not isinstance(dict_md, dict):
            raise TypeError('Metadata should be given as a dictionary '
                            'name: dict of metadata')
        self.dmd = dict_md
        self.exclude = exclude
        self.all_md = self.build_metadata_dict()

    def build_metadata_dict(self):
        '''Build the dictionary of metadata.

        Contains all the metadata for all samples.
        '''
        all_keys = (set(key for md in self.dmd.values() for key in md)
                    .difference(self.exclude))
        # we could write the whole following loop as a nested dict
        # comprehension, but let's not
        cdict = {}
        for key in all_keys:
            cdict[key] = {name: tmd[key] if key in tmd else MISSING
                          for name, tmd in self.dmd.items()}
        return cdict

    def compare_metadata(self):
        '''Metadata are compared with respect to the first one.'''
        bdict = {}
        t0name = sorted(self.dmd)[0]
        for key, kdict in self.all_md.items():
            bdict[key] = {n: v == kdict[t0name] for n, v in kdict.items()}
        return bdict

    def evaluate(self):
        '''Evaluate this test and turn it into a :class:`TestResultMetadata`.
        '''
        return TestResultMetadata(self, self.compare_metadata())

    def data(self):
        '''Generator yielding objects supporting the buffer protocol that (as a
        whole) represent a serialized version of `self`.'''
        yield from super().data()
        yield self.__class__.__name__.encode('utf-8')
        for key, tmd in self.dmd.items():
            yield key.encode('utf-8')
            for mdkey, mdval in tmd.items():
                yield mdkey.encode('utf-8')
                yield str(mdval).encode('utf-8')
        for key in self.exclude:
            yield key.encode('utf-8')



# Copyright French Alternative Energies and Atomic Energy Commission
# Contributors: valjean developers
# valjean-support@cea.fr
#
# This software is a computer program whose purpose is to analyze and
# post-process numerical simulation results.
#
# This software is governed by the CeCILL license under French law and abiding
# by the rules of distribution of free software. You can use, modify and/ or
# redistribute the software under the terms of the CeCILL license as circulated
# by CEA, CNRS and INRIA at the following URL: http://www.cecill.info.
#
# As a counterpart to the access to the source code and rights to copy, modify
# and redistribute granted by the license, users are provided only with a
# limited warranty and the software's author, the holder of the economic
# rights, and the successive licensors have only limited liability.
#
# In this respect, the user's attention is drawn to the risks associated with
# loading, using, modifying and/or developing or reproducing the software by
# the user in light of its specific status of free software, that may mean that
# it is complicated to manipulate, and that also therefore means that it is
# reserved for developers and experienced professionals having in-depth
# computer knowledge. Users are therefore encouraged to load and test the
# software's suitability as regards their requirements in conditions enabling
# the security of their systems and/or data to be ensured and, more generally,
# to use and operate it in the same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

'''This module defines a few useful functions and classes to embed generic
command execution in a :mod:`~.depgraph`.

Spawning external processes
===========================

.. doctest:: code
   :hide:

   >>> from valjean.config import Config
   >>> config = Config()

The :class:`RunTask` class is the basic building block to run tasks that
consist in spawning external processes and waiting for their completion.  It
makes it possible to execute arbitrary commands. Consider:

   >>> from valjean.cosette.run import RunTask
   >>> task = RunTask.from_cli(name='say', cli=['echo', 'ni!'])
   >>> env_update, status = task.do(env=dict(), config=config) # prints 'ni!'
   >>> print(status)
   TaskStatus.DONE

The task succeeded, but where is the output of our command?! Note that
:class:`RunTask` captures standard output and standard error and redirects them
to files. If you want to see what was printed, you have to look there:

   >>> def print_stdout(env_up, name):
   ...   """A small function to print the stdout of a task."""
   ...   stdout = env_up[name]['stdout']
   ...   with open(stdout) as stdout_f:
   ...     print(stdout_f.read(), end='')
   >>> print_stdout(env_update, 'say')
   ni!

Note that `command` is not parsed by a shell. So the following may not do what
you expect:

   >>> task = RunTask.from_cli(name='want',
   ...                         cli=['echo', 'We want... ', '&&',
   ...                              'echo', 'a shrubbery!'])
   >>> env_update, status = task.do(env=dict(), config=config)
   >>> print_stdout(env_update, 'want')
   We want...  && echo a shrubbery!

Indeed, :class:`RunTask` ran :command:`echo` only once.  If you need to execute
several commands, you can wrap them in a shell script and execute it.
Alternatively, you can directly invoke the :meth:`RunTask.from_clis` class
method:

   >>> task = RunTask.from_clis(name='want',
   ...                          clis=[['echo', '-n', 'We want... '],
   ...                                ['echo', 'a shrubbery!']])
   >>> env_update, status = task.do(env=dict(), config=config)
   >>> print_stdout(env_update, 'want')
   We want... a shrubbery!


Creating tasks using a factory
==============================

When you want to create multiple :class:`RunTask` objects using the same
executable, it can be convenient to use :class:`RunTaskFactory`. This class can
be parametrized to create tasks by specifying the path to the executable once
and for all, for instance, and providing the missing arguments later.

.. doctest:: RunTaskFactory
    :hide:

    >>> import os
    >>> from valjean.config import Config
    >>> config = Config()
    >>> def print_stdout(env_up, name):
    ...   """A small function to print the stdout of a task."""
    ...   stdout = env_up[name]['stdout']
    ...   with open(stdout) as stdout_f:
    ...     print(stdout_f.read(), end='')

The simplest way to create a :class:`RunTaskFactory` is to use one of the
:meth:`RunTaskFactory.from_executable` or :meth:`RunTaskFactory.from_task`
class methods. For example, this will create a factory instance that generates
:class:`RunTask` objects for the :command:`echo` executable:

    >>> factory = RunTaskFactory.from_executable('echo', name='echo')

You can use it to generate tasks by invoking the :meth:`RunTaskFactory.make`
method:

    >>> task = factory.make(name='task', extra_args=['spam'])

This creates an `task.echo` object (of type :class:`RunTask`) that executes
:file:`echo spam` when run:

    >>> env_up, status = task.do(env={}, config=config)
    >>> print_stdout(env_up, 'task.echo')
    spam

You can also leave the `name` parameter out. If you do so,
:class:`RunTaskFactory` will generate a name for you:

    >>> task_sausage = factory.make(extra_args=['sausage'])
    >>> task_sausage.name
    '....echo'

Of course you can generate multiple tasks using the same factory (this is the
whole point of :class:`RunTaskFactory`, really):

    >>> task_spam = factory.make(name='task_spam', extra_args=['spam'])
    >>> task_eggs = factory.make(name='task_eggs', extra_args=['eggs'])
    >>> task_bacon = factory.make(name='task_bacon', extra_args=['bacon'])

You can also specify a few arguments beforehand and provide the rest later:

    >>> factory = RunTaskFactory.from_executable('echo', name='echo',
    ...                                          default_args=['spam'])
    >>> task = factory.make(name='task', extra_args=['eggs'])
    >>> env_up, status = task.do(env={}, config=config)
    >>> print_stdout(env_up, 'task.echo')
    spam eggs

Finally, you can parametrize your arguments on arbitrary keywords that will be
provided when the task is created:

    >>> args = ['{food}', 'with', '{side}']
    >>> factory = RunTaskFactory.from_executable('echo', name='echo',
    ...                                          default_args=args)
    >>> task = factory.make(name='task', food='lobster', side='spam')
    >>> env_up, status = task.do(env={}, config=config)
    >>> print_stdout(env_up, 'task.echo')
    lobster with spam

Default values for the parameters may be specified when creating the factory
and can be overridden when the task is created:

    >>> args = ['{food}', 'with', '{side}']
    >>> factory = RunTaskFactory.from_executable('echo', default_args=args,
    ...                                          side='spam', name='echo')
    >>> beans = factory.make(name='baked beans', food='baked beans')
    >>> eggs = factory.make(name='eggs', food='eggs',
    ...                     side='bacon and spam')
    >>> env_up, status = beans.do(env={}, config=config)
    >>> print_stdout(env_up, 'baked beans.echo')
    baked beans with spam
    >>> env_up, status = eggs.do(env={}, config=config)
    >>> print_stdout(env_up, 'eggs.echo')
    eggs with bacon and spam

Note also that you can refer to the environment or the configuration in your
command-line arguments:

    >>> args = ['{env[side]}']
    >>> factory = RunTaskFactory.from_executable('echo', name='echo',
    ...                                          default_args=args)
    >>> task = factory.make(name='task')
    >>> env_up, status = task.do(env={'side': 'spam'}, config=config)
    >>> print_stdout(env_up, 'task.echo')
    spam


Caching
-------

The :class:`RunTaskFactory` class caches generated tasks under the hood.
Repeated calls to :meth:`RunTaskFactory.make` from the same factory with the
same arguments will result in the same task:

    >>> factory = RunTaskFactory.from_executable('echo', name='echo')
    >>> task_sausage = factory.make(extra_args=['sausage'])
    >>> task_sausage_again = factory.make(extra_args=['sausage'])
    >>> task_sausage is task_sausage_again
    True

If you instantiate another factory, the caching mechanism is defeated:

    >>> other_factory = RunTaskFactory.from_executable('echo', name='echo')
    >>> task_sausage_other = other_factory.make(extra_args=['sausage'])
    >>> task_sausage is task_sausage_other
    False


Module API
==========
'''

import os
import shlex
import logging
from functools import partial
from subprocess import call

from ..chrono import Chrono
from ..path import ensure, sanitize_filename
from .task import TaskStatus, det_hash
from .pythontask import PythonTask


LOGGER = logging.getLogger(__name__)


# The following functions are helpers for RunTask, but they may be used
# elsewhere, too
def run(clis, stdout, stderr, **subprocess_args):
    '''Run the given command lines and capture their stdout/stderr.

    Execution stops at the first failure (result value != 0).

    :param list clis: The list of commands to execute.
    :param stdout: File handle to capture the stdout stream.
    :type stdout: :term:`file object`
    :param stderr: File handle to capture the stderr stream.
    :type stderr: :term:`file object`
    :param dict subprocess_args: Parameters to be passed to
        :func:`subprocess.call`.
    '''
    results = []    # collect the return codes of each cli
    status = TaskStatus.DONE    # will change to FAILED in case of trouble
    with Chrono() as chrono:
        for cli in clis:
            LOGGER.debug('Running cli: %s', cli)
            print('$ ' + ' '.join(shlex.quote(token) for token in cli),
                  file=stderr, flush=True)
            LOGGER.debug('subprocess_args: %s', subprocess_args)
            result = call(cli, universal_newlines=True,
                          stdout=stdout, stderr=stderr,
                          **subprocess_args)
            LOGGER.debug('Run result: %s', result)
            results.append(result)
            if result != 0:
                LOGGER.warning('A subprocess ended with return code %s; will '
                               'skip the remaining commands', result)
                status = TaskStatus.FAILED
                break
    return results, status, float(chrono)


def make_cap_paths(base_path):
    '''Construct filenames to capture stdout and stderr.'''
    stdout_path = base_path / 'stdout'
    stderr_path = base_path / 'stderr'
    ensure(stdout_path)
    ensure(stderr_path)
    return stdout_path.resolve(), stderr_path.resolve()


class RunTask(PythonTask):
    '''Task that executes the specified shell commands and waits for their
    completion.'''

    @classmethod
    def from_cli(cls, name, cli, **kwargs):
        '''Create a :class:`RunTask` from a single command line.

        Use the :meth:`RunTask.from_clis` method if you want to run several
        commands in a row.

        :param str name: The name of this task.
        :param list cli: The command line to be executed, as a list of strings.
            The first element of the list is the command and the following ones
            are its arguments.
        :param kwargs: Any other keyword arguments will be passed on to the
            constructor (see :meth:`RunTask.__init__`).
        '''
        if not isinstance(cli, (list, tuple)):
            raise TypeError('Argument `cli` must be a list or a tuple')
        return cls.from_clis(name, [cli], **kwargs)

    @classmethod
    def from_clis(cls, name, clis, **kwargs):
        '''Create a :class:`RunTask` from a list of command lines.

        :param str name: The name of this task.
        :param list clis: The command lines to be executed, as a list of lists
            of strings.  The first element of each sub-list is the command and
            the following ones are its arguments.
        :param kwargs: Any other keyword arguments will be passed on to the
            constructor (see :meth:`RunTask.__init__`).
        '''
        if (not isinstance(clis, (list, tuple))
                or not all(isinstance(cli, (list, tuple)) for cli in clis)):
            raise TypeError('Argument `cli` must be a list (or a tuple) '
                            'of lists (or tuples)')
        return cls(name, lambda _env, _config: clis, **kwargs)

    def __init__(self, name, clis_closure, *, deps=None, soft_deps=None,
                 **subprocess_args):
        '''Initialize this task from a list of command lines.

        The `clis_closure` argument must be a closure. It will be invoked at
        execution time as::

            clis_closure(env, config)

        and it must return the command lines to be executed, as a list of lists
        of strings.

        :param str name: The name of this task.
        :param list clis_closure: A closure to generate the command lines.
        :param dict subprocess_args: Any remaining options will be passed to
            the :class:`.subprocess.Popen` constructor.
        :param deps: The dependencies for this task (see
            :meth:`Task.__init__ <valjean.cosette.task.Task.__init__>` for the
            format), or `None`.
        :type deps: list(Task) or None
        :param soft_deps: The dependencies for this task (see
            :meth:`Task.__init__ <valjean.cosette.task.Task.__init__>` for the
            format), or `None`.
        :type soft_deps: list(Task) or None
        '''
        super().__init__(name, self.run_task(clis_closure, name,
                                             **subprocess_args),
                         deps=deps, soft_deps=soft_deps,
                         env_kwarg='env', config_kwarg='config')
        LOGGER.debug('Created %s task %r', self.__class__.__name__, self.name)
        LOGGER.debug('  - deps = %s', deps)
        LOGGER.debug('  - soft_deps = %s', soft_deps)
        LOGGER.debug('  - subprocess_args = %s', subprocess_args)

    def run_task(self, clis_closure, name, **subprocess_args):
        # pylint: disable=unused-argument
        '''Execute the specified command and wait for its completion.

        On completion, this method proposes the following updates to the
        environment::

            env[task.name]['clis'] = clis
            env[task.name]['return_codes'] = return_codes
            env[task.name]['elapsed_time'] = wallclock_time
            env[task.name]['stdout'] = stdout
            env[task.name]['stderr'] = stderr

        Here ``clis`` is the list of command lines that were executed,
        ``return_codes`` is the list of return codes of the executed commands,
        and ``elapsed_time`` is the time the whole list took. The keys
        ``stdout`` and ``stderr`` hold the paths to the files containing
        respectively the captured standard output and standard error streams.

        :param Env env: The task environment.
        :param config: The configuration object.
        :type config: Config or None
        :returns: The proposed environment update.
        '''
        def runner(*, env, config, name, clis_closure, subprocess_args):
            # pylint: disable=too-many-locals
            '''Actually run the command lines.'''
            from pathlib import Path

            clis = clis_closure(env, config)
            output_dir = Path(config.query('path', 'output-root'),
                              sanitize_filename(name))
            stdout_path, stderr_path = make_cap_paths(output_dir)

            with stdout_path.open('w') as stdout:
                with stderr_path.open('w') as stderr:
                    results, status, elapsed = run(clis, stdout, stderr,
                                                   cwd=str(output_dir),
                                                   **subprocess_args)

            env_up = {self.name: {'clis': clis,
                                  'output_dir': str(output_dir),
                                  'return_codes': results,
                                  'elapsed_time': elapsed,
                                  'result': str(stdout_path),
                                  'stdout': str(stdout_path),
                                  'stderr': str(stderr_path)}}

            LOGGER.debug('RunTask %s ends, elapsed time: %f s', name, elapsed)
            return env_up, status

        # saturate name, clis and subprocess_args with the values from the
        # enclosing scope
        return partial(runner, name=name, clis_closure=clis_closure,
                       subprocess_args=subprocess_args)


class RunTaskFactory:
    '''Create multiple tasks from the same executable without even breaking a
    sweat.'''

    @classmethod
    def _generic_clis_closure(cls, executable, *, env, config, default_args,
                              extra_args, **kwargs):
        # pylint: disable=too-many-arguments
        cli = [executable]
        cli.extend(arg.format(env=env, config=config, **kwargs)
                   for arg in default_args)
        cli.extend(extra_args)
        return [cli]

    @classmethod
    def from_task(cls, task, *, relative_path, name=None, default_args=None,
                  **kwargs):
        '''This class method creates a :class:`RunTaskFactory` from a
        :class:`~valjean.cosette.task.Task`. The command to be executed
        must appear in the `output_dir` directory of the given task; its
        relative location must be specified with the `relative_path` argument.

        This method can be used, among other things, to create a
        :class:`RunTaskFactory` from a
        :class:`~valjean.cosette.code.CheckoutTask` or a
        :class:`~valjean.cosette.code.BuildTask`.

        :param task: The :class:`~valjean.cosette.task.Task` object.
        :type task: :class:`~valjean.cosette.task.Task`
        :param str relative_path: The path to the executable, relative to the
            the output directory (`output_dir`) of the task.
        :param name: a unique identifier for this factory. The factory name is
            used to produce unique names for the generated tasks.  If `None` is
            given, the factory name will be constructed by hashing the other
            arguments.
        :type name: str or None
        :param default_args: see :meth:`RunTaskFactory.from_executable`.
        :param kwargs: Any remaining arguments will be passed to the
            :meth:`__init__`.
        '''
        d_args = [] if default_args is None else default_args
        LOGGER.debug('relative_path=%r', relative_path)
        LOGGER.debug('d_args=%s', d_args)
        closure = partial(cls._clis_closure_from_task, task.name,
                          relative_path, d_args)
        name = (name if name is not None
                else det_hash(task.name, relative_path, d_args))
        return cls(closure, deps=[task], name=name, **kwargs)

    @classmethod
    def _clis_closure_from_task(cls, job_name, relative_path, default_args,
                                extra_args, **kwargs):
        '''This static method generates the closure consumed by
        :meth:`RunTaskFactory.from_task`.'''
        def clis_closure(env, config):
            output_dir = env[job_name]['output_dir']
            executable = os.path.join(output_dir, relative_path)
            return cls._generic_clis_closure(executable, env=env,
                                             config=config,
                                             default_args=default_args,
                                             extra_args=extra_args,
                                             **kwargs)
        return clis_closure

    @classmethod
    def from_executable(cls, path, name=None, default_args=None, **kwargs):
        '''This class method creates a :class:`RunTaskFactory` from the path to
        an existing executable.

        :param str path: the path to the executable.
        :param name: a unique identifier for this factory. If `None` is given,
            the factory will use a hash of the other arguments.
        :type name: str or None
        :param default_args: The list of arguments that will be passed to the
            executed command by :class:`RunTask`. It may contain expressions
            understood by Python's format mini-language, such as ``{foo}``.
            These expressions can be filled when the task is generated by
            passing appropriate keyword arguments to the :meth:`make` method.
        :type default_args: list(str) or None
        :param kwargs: Any remaining arguments will be passed to the
            :meth:`__init__`.
        '''
        d_args = [] if default_args is None else default_args
        LOGGER.debug('path=%r', path)
        LOGGER.debug('d_args=%s', d_args)
        name = name if name is not None else det_hash(path, d_args)
        return cls(partial(cls._clis_closure_from_executable, path, d_args),
                   name=name, **kwargs)

    @classmethod
    def _clis_closure_from_executable(cls, executable, default_args,
                                      extra_args, **kwargs):
        def clis_closure(env, config):
            return cls._generic_clis_closure(executable, env=env,
                                             config=config,
                                             default_args=default_args,
                                             extra_args=extra_args,
                                             **kwargs)
        return clis_closure

    def __init__(self, make_closure, *, deps=None, soft_deps=None, name,
                 **kwargs):
        self.make_closure = make_closure
        self.deps = [] if deps is None else deps
        self.soft_deps = [] if soft_deps is None else soft_deps
        self.kwargs = kwargs
        self.cache = {}
        self.name = name
        LOGGER.debug('creating factory with name: %s', self.name)

    def make(self, *, name=None, extra_args=None,
             subprocess_args=None, deps=None, soft_deps=None, **kwargs):
        '''Create a :class:`RunTask` object.

        :param name: the name of the task to be generated, as a string. If
            `None` is passed, :class:`RunTaskFactory` will generate a name by
            hashing the contents of all the other arguments.
        :type name: str or None
        :param extra_args: A list of additional arguments that will be appended
            at the end of the command line.
        :type extra_args: list or None
        :param subprocess_args: A dictionary of arguments for the call to the
            :class:`.subprocess.Popen` constructor.
        :type subprocess_args: dict or None
        :param deps: A list of dependencies for the generated task. Note that
            factories built with :meth:`from_task` automatically inject
            dependencies on the given task.
        :type deps: list(Task) or None
        :param soft_deps: A list of soft dependencies for the generated task.
        :type soft_deps: list(Task) or None
        :param kwargs: Any remaining keyword arguments will be used to format
            the command line before execution. The environment and the
            configuration are available at formatting time as ``env`` and
            ``config``, respectively.
        '''
        LOGGER.debug('RunFactory.make(name=%r, extra_args=%r, '
                     'subprocess_args=%r, deps=%r, soft_deps=%r, **kwargs=%r)',
                     name, extra_args, subprocess_args, deps, soft_deps,
                     kwargs)
        extra_args = [] if extra_args is None else extra_args
        subprocess_args = {} if subprocess_args is None else subprocess_args
        deps = [] if deps is None else deps
        soft_deps = [] if soft_deps is None else soft_deps
        kwargs_ = self.kwargs.copy()
        kwargs_.update(kwargs)
        cli_closure = self.make_closure(extra_args, **kwargs_)

        # handle caching
        if name is None:
            task_name = str(det_hash(self.name, extra_args, kwargs_))
        else:
            task_name = name
        task_name += '.' + self.name
        if task_name in self.cache:
            LOGGER.debug('cache hit for task name %r', task_name)
            cached_task = self.cache[task_name]
            return cached_task

        LOGGER.debug('cache miss for task name %r', task_name)

        task = RunTask(task_name, cli_closure,
                       deps=self.deps + deps,
                       soft_deps=self.soft_deps + soft_deps,
                       **subprocess_args)
        self.cache[task_name] = task
        return task

    def copy(self):
        '''Return a copy of this object.'''
        return self.__class__(self.make_closure,
                              deps=self.deps.copy(),
                              soft_deps=self.soft_deps.copy(),
                              name=self.name, **self.kwargs)



# Copyright French Alternative Energies and Atomic Energy Commission
# Contributors: valjean developers
# valjean-support@cea.fr
#
# This software is a computer program whose purpose is to analyze and
# post-process numerical simulation results.
#
# This software is governed by the CeCILL license under French law and abiding
# by the rules of distribution of free software. You can use, modify and/ or
# redistribute the software under the terms of the CeCILL license as circulated
# by CEA, CNRS and INRIA at the following URL: http://www.cecill.info.
#
# As a counterpart to the access to the source code and rights to copy, modify
# and redistribute granted by the license, users are provided only with a
# limited warranty and the software's author, the holder of the economic
# rights, and the successive licensors have only limited liability.
#
# In this respect, the user's attention is drawn to the risks associated with
# loading, using, modifying and/or developing or reproducing the software by
# the user in light of its specific status of free software, that may mean that
# it is complicated to manipulate, and that also therefore means that it is
# reserved for developers and experienced professionals having in-depth
# computer knowledge. Users are therefore encouraged to load and test the
# software's suitability as regards their requirements in conditions enabling
# the security of their systems and/or data to be ensured and, more generally,
# to use and operate it in the same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

'''This module extends the functionalities provided by :mod:`~.cosette.use` to
the commonplace task of parsing Tripoli-4 output files. The decorators defined
by this module, :func:`using_parse_result` and :func:`using_browser`,
can be used to remove boilerplate code that parses and accesses the Tripoli-4
results. At the same time, they simplify the integration of user-defined
post-processing (tests) into the :mod:`valjean` workflow.

.. doctest:: eponine_use
    :hide:

    >>> # pylint: disable=trailing-whitespace,line-too-long
    >>> # noqa: E501,W291
    >>> from valjean.config import Config
    >>> config = Config()

We will need to set up a few things to show an example of how to use the
decorators. We don't want to build and run Tripoli-4 for the purpose of this
documentation, but we can mock a Tripoli-4 run by writing an existing Tripoli-4
output to a file. Here is the sample output::

    >>> example = """BATCH 10
    ... initialization time (s): 7
    ...  batch number : 10
    ... *********************************************************
    ...  RESULTS ARE GIVEN FOR SOURCE INTENSITY : 8.111452e+04
    ... *********************************************************
    ...
    ...
    ...  Mean weight leakage = 7.929746e+00	 sigma = 7.929746e+00	 sigma% = 1.000000e+02
    ...
    ...
    ...  Edition after batch number : 10
    ...
    ...
    ...
    ... ******************************************************************************
    ... RESPONSE FUNCTION : FLUX
    ... RESPONSE NAME :
    ... ENERGY DECOUPAGE NAME : DEC_SPECTRE
    ...
    ...
    ...  PARTICULE : NEUTRON
    ... ******************************************************************************
    ...
    ... 	 scoring mode : SCORE_SURF
    ... 	 scoring zone : 	 Frontier 	 volumes : 14,27
    ...
    ...
    ... 	 SPECTRUM RESULTS
    ... 	 number of first discarded batches : 0
    ...
    ... 	 group			 score		 sigma_% 	 score/lethargy
    ... Units:	 MeV			 neut.s^-1	 %		 neut.s^-1
    ...
    ... 2.000000e+01 - 1.000000e-11	1.307419e+00	8.719708e+01	1.046074e+01
    ...
    ... 	 ENERGY INTEGRATED RESULTS
    ...
    ... 	 number of first discarded batches : 0
    ...
    ... number of batches used: 10	1.307419e+00	8.719708e+01
    ...
    ...
    ...
    ...  simulation time (s) : 20
    ...
    ...
    ...  Type and parameters of random generator at the end of simulation:
    ... 	 DRAND48_RANDOM 13531 45249 20024  COUNTER	2062560
    ...
    ...
    ...  simulation time (s): 20
    ...
    ...
    ... =====================================================================
    ... 	NORMAL COMPLETION
    ... =====================================================================
    ... """

We construct a :class:`~.RunTaskFactory` that simply echoes any text passed as
an argument::

    >>> from valjean.cosette.run import RunTaskFactory
    >>> echo_factory = RunTaskFactory.from_executable('/bin/echo', name='echo',
    ...                                               default_args=['{text}'])

In the real world, the :class:`~.RunTaskFactory` would actually generate tasks
to run a Tripoli-4 executable.


Injecting the raw parse results
===============================

Now we can construct the decorator that parses the Tripoli-4 results and apply
it to a Python function::

    >>> from valjean.eponine.tripoli4 import use
    >>> using_last_parse_result = use.using_last_parse_result(echo_factory)
    >>> @using_last_parse_result(text=example)
    ... def source(results):
    ...     return results.res['batch_data']['source_intensity']

If we inspect `source`, we can see that it is a :class:`~.Use` object::

    >>> type(source)
    <class 'valjean.cosette.use.Use'>

If you are not sure what :class:`~.Use` does, this is a good moment to go and
read its documentation. Don't worry, I'll wait.

*(time passes)*

:mod:`valjean` can inspect this object and convert it into a :class:`~.Task`::

    >>> source_task = source.get_task()
    >>> type(source_task)
    <class 'valjean.cosette.pythontask.PythonTask'>
    >>> source_task.name
    '....source'

The task has explicit dependencies, which means that it will be correctly
integrated in the :mod:`valjean` dependency graph. In particular, this task
depends on a task that actually does the parsing::

    >>> source_task.depends_on
    {Task('....parse_batch_index')}
    >>> parse_task = next(iter(source_task.depends_on))

It depends on the `make_parser_task`, that scans the file and build the parser

    >>> parse_task.depends_on
    {Task('....make_parser')}
    >>> make_parser_task = next(iter(parse_task.depends_on))

and `make_parser_task` in turn depends on the :class:`~.RunTask` that was
generated by `echo_factory`::

    >>> make_parser_task.depends_on
    {Task('....echo')}
    >>> run_task = next(iter(make_parser_task.depends_on))

We manually execute the tasks in the correct order and we check that we recover
the right result at the end::

    >>> from valjean.cosette.env import Env
    >>> env = Env()
    >>> for task in [run_task, make_parser_task, parse_task, source_task]:
    ...     env_up, _ = task.do(env=env, config=config)
    ...     env.apply(env_up)
    >>> print(env[source_task.name]['result'])
    81114.52


Using a :class:`~.browser.Browser`
=============================================

The raw parse results are useful in some situations, but most of the time you
probably want to work with a higher-level representation of the calculation
result, such as a :class:`~.browser.Browser`. This module provides a function
to construct a decorator that automatically creates the :class:`~.RunTask` from
the :class:`~.RunTaskFactory`, runs the task, parses the resulting output and
wraps the parse results in an :class:`~.browser.Browser`::


    >>> using_browser = use.using_browser(echo_factory, 10)
    >>> @using_browser(text=example)
    ... def extract_simulation_time(browser):
    ...     return browser.globals['simulation_time']

We can again check that everything went as expected by manually unwrapping the
sequence of tasks and running them::

    >>> extract_task = extract_simulation_time.get_task()
    >>> browse_task = next(iter(extract_task.depends_on))
    >>> parse_task = next(iter(browse_task.depends_on))
    >>> make_parser_task = next(iter(parse_task.depends_on))
    >>> run_task = next(iter(make_parser_task.depends_on))
    >>> env = Env()
    >>> for task in [run_task, make_parser_task, parse_task,
    ...              browse_task, extract_task]:
    ...     env_up, _ = task.do(env=env, config=config)
    ...     env.apply(env_up)
    >>> print(env[extract_task.name]['result'])
    20


Module API
==========
'''

import functools
from .parse import Parser
from ...cosette.use import UseRun
from ...cosette.pythontask import TaskException
from ..tripoli4.parse import ParserException


def partial(func, *args, **kwargs):
    '''An improved version of :func:`functools.partial` that calls
    :func:`functools.update_wrapper` on the partially applied function in order
    to update its metadata (name, etc.).'''
    partial_func = functools.partial(func, *args, **kwargs)
    functools.update_wrapper(partial_func, func)
    return partial_func


def make_parser(filename):
    '''Create a Parser object and scan a Tripoli-4 output file.

    :param str filename: the name of the file to parse.
    :param int batch: the number of the batch to parse; see
                      :meth:`~.Parser.__init__`.
    :raises ValueError: if parsing fails.
    :returns: the parser
    :rtype: Parser
    '''
    try:
        parser = Parser(filename)
    except ParserException as tpe:
        raise TaskException(f'cannot build parser {filename}: {tpe}') from None
    return parser


def parse_batch_number(parser, *, batch_number):
    '''Parse a batch result from Tripoli-4.

    :param int batch_number: batch number
    :rtype: ParseResult
    '''
    try:
        pres = parser.parse_from_number(batch_number=batch_number)
    except ParserException as t4pe:
        raise TaskException(f'cannot parse {parser.jdd}: {t4pe}') from None
    return pres


def parse_batch_index(parser, *, batch_index=-1):
    '''Parse a batch result from Tripoli-4.

    :param int batch_index: index of the batch in the list of batches
    :rtype: ParseResult
    '''
    try:
        pres = parser.parse_from_index(batch_index=batch_index)
    except ParserException as t4pe:
        raise TaskException(f'cannot parse {parser.jdd}: {t4pe}') from None
    return pres


def using_parser(factory):
    '''Construct a decorator that injects Tripoli-4 parser into a Python
    function.

    :param factory: a factory producing Tripoli-4 runs.
    :type factory: :class:`~.RunTaskFactory`
    :returns: a decorator (see the module docstring for more information).
    '''
    use_run = UseRun.from_factory(factory)
    return use_run.map(make_parser)


def using_parse_result(factory, batch_number):
    '''Construct a decorator that injects the raw Tripoli-4 parse results into
    a Python function.

    :param factory: a factory producing Tripoli-4 runs.
    :type factory: :class:`~.RunTaskFactory`
    :param int batch_number: the number of the batch to parse; see
        :meth:`~.Parser.__init__`.
    :returns: a decorator (see the module docstring for more information).
    '''
    use_parser = using_parser(factory)
    return use_parser.map(partial(parse_batch_number,
                                  batch_number=batch_number))


def using_last_parse_result(factory):
    '''Construct a decorator that injects the last raw Tripoli-4 parse results
    into a Python function.

    :param factory: a factory producing Tripoli-4 runs.
    :type factory: :class:`~.RunTaskFactory`
    :returns: a decorator (see the module docstring for more information).
    '''
    use_parser = using_parser(factory)
    return use_parser.map(partial(parse_batch_index, batch_index=-1))


def to_browser(result):
    '''Create a :class:`~.Browser` from the parsing result.

    :param ParseResult result: result from T4 parser
    :returns: the browser
    :rtype: Browser
    '''
    return result.to_browser()


def using_browser(factory, batch_number):
    '''Construct a decorator that injects an
    :class:`~.browser.Browser` to the Tripoli-4 parse results into a
    Python function.

    :param factory: a factory producing Tripoli-4 runs.
    :type factory: :class:`~.RunTaskFactory`
    :param int batch_number: the number of the required batch (will be parsed
        then transformed in  :class:`~.browser.Browser`)
    :returns: a decorator (see the module docstring for more information).
    '''
    use_parse_result = using_parse_result(factory, batch_number=batch_number)
    return use_parse_result.map(to_browser)


def using_last_browser(factory):
    '''Construct a decorator that injects a
    :class:`~.browser.Browser` to the Tripoli-4 last parse results
    into a Python function.

    :param factory: a factory producing Tripoli-4 runs.
    :type factory: :class:`~.RunTaskFactory`
    :returns: a decorator (see the module docstring for more information).
    '''
    use_parse_result = using_last_parse_result(factory)
    return use_parse_result.map(to_browser)



from functools import update_wrapper

class LoopMarker:
    def __init__(self, values):
        self.values = tuple(values)

    def __call__(self, var):
        return Loop(var, self.values)

class Loop:
    def __init__(self, var, values):
        self.var = var
        self.values = values

    def __call__(self, wrapped):
        return LoopDecorator(wrapped, self.var, self.values)

class LoopDecorator:
    def __init__(self, wrapped, var, values):
        self.var = var
        self.values = values
        update_wrapper(self, wrapped, updated=())

    def __call__(self, *args, **kwargs):
        '''Redirect the call to the wrapped function.'''
        # pylint: disable=no-member
        return self.__wrapped__(*args, **kwargs)

    def __repr__(self):
        '''Redirect the call to the wrapped function.'''
        # pylint: disable=no-member
        return repr(self.__wrapped__)

    def get_tasks(self):
        for value in self.values:
            pass



# Copyright French Alternative Energies and Atomic Energy Commission
# Contributors: valjean developers
# valjean-support@cea.fr
#
# This software is a computer program whose purpose is to analyze and
# post-process numerical simulation results.
#
# This software is governed by the CeCILL license under French law and abiding
# by the rules of distribution of free software. You can use, modify and/ or
# redistribute the software under the terms of the CeCILL license as circulated
# by CEA, CNRS and INRIA at the following URL: http://www.cecill.info.
#
# As a counterpart to the access to the source code and rights to copy, modify
# and redistribute granted by the license, users are provided only with a
# limited warranty and the software's author, the holder of the economic
# rights, and the successive licensors have only limited liability.
#
# In this respect, the user's attention is drawn to the risks associated with
# loading, using, modifying and/or developing or reproducing the software by
# the user in light of its specific status of free software, that may mean that
# it is complicated to manipulate, and that also therefore means that it is
# reserved for developers and experienced professionals having in-depth
# computer knowledge. Users are therefore encouraged to load and test the
# software's suitability as regards their requirements in conditions enabling
# the security of their systems and/or data to be ensured and, more generally,
# to use and operate it in the same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

'''This module provides the :class:`~.Chrono` context manager.'''

import time


class Chrono:
    '''Time some code and store the elapsed time.

    This class is a simple context manager that measures the execution time of
    a code fragment. The elapsed time is stored as a float in `self.elapsed`
    and can be accessed as such.

    Examples:

    >>> with Chrono() as chrono:
    ...     print(42)
    42
    >>> print(f'printing 42 took {chrono} seconds')
    printing 42 took ... seconds
    '''

    def __init__(self):
        self._start = None
        self.elapsed = None

    def __enter__(self):
        self._start = time.perf_counter()
        return self

    def __exit__(self, _exc_type, _exc_value, _traceback):
        self.elapsed = time.perf_counter() - self._start
        self._start = None

    def __str__(self):
        '''Return the elapsed time, as a string.'''
        return str(self.elapsed)

    def __float__(self):
        '''Return the elapsed time, as a float.'''
        return float(self.elapsed)

    def __int__(self):
        '''Return the elapsed time, as an int.'''
        return int(self.elapsed)

    def __format__(self, format_spec):
        '''Return the elapsed time, formatted according to `format_spec`.

        >>> chrono = Chrono()
        >>> chrono.elapsed = 1e-6
        >>> f'{chrono}'
        '1e-06'
        >>> f'{chrono:f}'
        '0.000001'
        '''
        return format(self.elapsed, format_spec)
